/*******************************************************************************

  PSPoste - Wireless Email Client for the Sony PSP
  Copyright (C) Benjamin Biron

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

  Module :	CBase.cpp

  Description :	

  Last Modified $Date: 2008/10/06 04:50:49 $

  $Revision: 1.2 $

*******************************************************************************/

#ifdef INSTRUMENTED_ALLOCS

//******************************************************************************
//   Include Files
//******************************************************************************
#include "CBase.h"

#include <psptypes.h>
#include <pspthreadman.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

//******************************************************************************
//   Local Macros
//******************************************************************************

//******************************************************************************
//   Local Constants
//******************************************************************************
#define ALLOC_HEADER_MAGIC 0xfa7beef

//******************************************************************************
//   Static Prototypes
//******************************************************************************

//******************************************************************************
//   Global Variables
//******************************************************************************

//******************************************************************************
//   Static Variables
//******************************************************************************
AllocHeaderList CBase::s_allocHeaderList;
SceUID CBase::s_listSema;
u32 CBase::s_nextAllocCount;

//******************************************************************************
//   Class Definition
//******************************************************************************

//******************************************************************************
//	
//******************************************************************************
AllocHeader::AllocHeader(int aBlockSize, int aAllocCount) :
	magic(ALLOC_HEADER_MAGIC),
	blockSize(aBlockSize),
	allocCount(aAllocCount),
	m_next(NULL),
	m_prev(NULL)
{
	
};

void AllocHeader::poison()
{
	memset(this, 0xba, sizeof(AllocHeader)+blockSize);
}

AllocHeaderList::AllocHeaderList() :
	m_head(NULL),
	m_tail(NULL)
{
	
};

void AllocHeaderList::append(AllocHeader* allocHeader)
{
	if (allocHeader->m_next != NULL || allocHeader->m_prev != NULL)
	{
		fprintf(stderr, "ALLOC ERROR: Inserting uninitialized alloc header\n");
	}
	
	if (m_tail)
	{
		/* First case: list isn't empty */
		m_tail->m_next = allocHeader;
		allocHeader->m_prev = m_tail;
		m_tail = allocHeader;
	}
	else
	{
		/* Second case: list is empty */
		m_head = allocHeader;
		m_tail = allocHeader;
	}
	
	if (m_head && (m_head->m_prev != NULL || m_head->magic != ALLOC_HEADER_MAGIC))
	{
		fprintf(stderr, "ALLOC ERROR: Alloc header list append corrupted the head of the list\n");
	}
	
	if (m_tail && (m_tail->m_next != NULL || m_tail->magic != ALLOC_HEADER_MAGIC))
	{
		fprintf(stderr, "ALLOC ERROR: Alloc header list append corrupted the tail of the list\n");
	}
}

void AllocHeaderList::remove(AllocHeader* allocHeader)
{
	if (allocHeader != m_head && allocHeader != m_tail)
	{
		/* First case: we're removing from somewhere in the middle of the list */
		allocHeader->m_prev->m_next = allocHeader->m_next;
		allocHeader->m_next->m_prev = allocHeader->m_prev;
	}
	else
	{
		if (allocHeader == m_head)
		{
			/* Second case: we're removing from the head of the list */
			m_head = allocHeader->m_next;
			if (m_head)
			{
				m_head->m_prev = NULL;
			}
		}
		
		if (allocHeader == m_tail)
		{
			/* Third case: we're removing from the tail of the list */
			m_tail = allocHeader->m_prev;
			if (m_tail)
			{
				m_tail->m_next = NULL;
			}
		}
	}
	
	if (m_head && (m_head->m_prev != NULL || m_head->magic != ALLOC_HEADER_MAGIC))
	{
		fprintf(stderr, "ALLOC ERROR: Alloc header list append corrupted the head of the list\n");
	}
	
	if (m_tail && (m_tail->m_next != NULL || m_tail->magic != ALLOC_HEADER_MAGIC))
	{
		fprintf(stderr, "ALLOC ERROR: Alloc header list remove corrupted the tail of the list\n");
	}
}

bool CBase::Open()
{
	s_listSema = sceKernelCreateSema("CBase_Sema", 0, 1, 1, NULL);
	new (&s_allocHeaderList) AllocHeaderList();
	s_nextAllocCount = 1;
	
	return true;
}

void CBase::Close()
{
	printLiveMemory();
	sceKernelDeleteSema(s_listSema);
	s_listSema = 0;
}

void* CBase::operator new(size_t size)
{
	AllocHeader* header = NULL;

	/* Allocate space for the header, and add it to the list */
	header = (AllocHeader*)malloc(size + sizeof(AllocHeader));
	lockList();
	/* BREAK HERE TO DEBUG LEAKS. Break with condition "s_nextAllocCount == N", where N == the allocation # */
	new (header) AllocHeader(size, s_nextAllocCount++);
	s_allocHeaderList.append(header);
	unlockList();
	
	/* Return the shifted pointer */
	return &header[1];
}

void CBase::operator delete(void *p)
{
	AllocHeader* header = (AllocHeader*)p;
	header = &header[-1];
	
	if (header->magic == ALLOC_HEADER_MAGIC)
	{
		lockList();
		s_allocHeaderList.remove(header);
		unlockList();
		header->poison();
		free(header);
	}
	else
	{
		fprintf(stderr, "ALLOC ERROR: Magic field doesn't match, likely freeing an invalid pointer\n");
	}
}

void CBase::printLiveMemory()
{
	lockList();
	int leakedBlockCount = 0;
	AllocHeader* cur = s_allocHeaderList.m_head;
	while (cur)
	{
		if (cur->magic != ALLOC_HEADER_MAGIC)
		{
			fprintf(stderr, "ALLOC ERROR: Corrupted alloc header list\n");
		}
		else
		{
			fprintf(stderr, "ALLOC WARNING: Leaked %d bytes due to allocation #%d\n", cur->blockSize, cur->allocCount);
		}
		leakedBlockCount += 1;
		cur = cur->m_next;
	}
	unlockList();
	
	if (leakedBlockCount)
	{
		fprintf(stderr, "ALLOC WARNING: Leaked %d blocks\n", leakedBlockCount);
	}
}

void CBase::lockList()
{
	sceKernelWaitSema(s_listSema, 1, NULL);
}

void CBase::unlockList()
{
	sceKernelSignalSema(s_listSema, 1);
}

#endif /* INSTRUMENTED_ALLOCS */
