/*******************************************************************************

  PSPoste - Wireless Email Client for the Sony PSP
  Copyright (C) Benjamin Biron, Sajeeth Cherian, David Cummings

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

  Module :	CConnectionManager.cpp

  Description :	

  Last Modified $Date: 2008/09/24 01:23:07 $

  $Revision: 1.9 $

*******************************************************************************/

//******************************************************************************
//   Include Files
//******************************************************************************
#include "CConnectionManager.h"
#include "CLog.h"

#include <pspsdk.h>
#include <pspnet.h>
#include <pspnet_inet.h>
#include <pspnet_apctl.h>
#include <pspnet_resolver.h>
#include <psputility.h>
#include <pspkerror.h>
#include <pspwlan.h>

//******************************************************************************
//   Local Macros
//******************************************************************************

//******************************************************************************
//   Local Constants
//******************************************************************************

//******************************************************************************
//   Static Prototypes
//******************************************************************************

//******************************************************************************
//   Global Variables
//******************************************************************************

//******************************************************************************
//   Static Variables
//******************************************************************************
int CConnectionManager::s_DefaultConnection;
unsigned int CConnectionManager::s_DefaultTimeout;
unsigned int CConnectionManager::s_DefaultPollingInterval;

int CConnectionManager::s_Connection;
EConnManState CConnectionManager::s_ConnManState;
int CConnectionManager::s_ConnectionTimeout;
int CConnectionManager::s_ErrFlag;

SceUID CConnectionManager::s_mbx;
SceUID CConnectionManager::s_sema;

bool CConnectionManager::s_DisconnectRequested;

std::vector<MPSPConnection> CConnectionManager::s_pspConnections;

//******************************************************************************
//   Class Definition
//******************************************************************************

//******************************************************************************
//	 Note the special handling of the disconnect case; we don't want to actually
//	 disconnect if there are still messages in the pipe. Ex: if a connect
//	 arrives, we don't want to disconnect then reconnect.
//******************************************************************************
int CConnectionManager::ConnManThreadCB(SceSize args, void* argp)
{
	s32 err;
	void *data;
	SConnManMsg* msg;

	s_mbx = sceKernelCreateMbx("ConnMan_Mailbox", 0, NULL);
	
	Initialize();
	
	while (1)
	{
		if (!s_DisconnectRequested)
		{
			if ((err = sceKernelReceiveMbx(s_mbx, &data, NULL)))
			{
				CLog::Error("sceKernelReceiveMbx returned %x; killing connection manager.\n", err);
				Close();
				return err;
			}
		}
		else
		{
			// We don't want to block here because if the pipe is empty, we want
			// to handle the actual disconnection
			err = sceKernelPollMbx(s_mbx, &data);
			if ((u32)err == SCE_KERNEL_ERROR_MBOX_NOMSG)
			{
				if ((err = Disconnect()))
					CLog::Error("CConnectionManager::Disconnect() error: 0x%x\n", err);
					
				continue;
			}
			else if (err < 0)
			{
				CLog::Error("sceKernelPollMbx returned %x; killing connection manager.\n", err);
				Close();
				return err;
			}
		}
		
		msg = reinterpret_cast<SConnManMsg*>(data);
		switch (msg->m_type)
		{
			case EConnManMsg_Connect:
				if ((err = Connect(msg->m_arg)))
					CLog::Error("CConnectionManager::Connect() error: 0x%x\n", err);
				break;
			case EConnManMsg_Disconnect:
				s_DisconnectRequested = true;
				break;
			case EConnManMsg_SetConn:
				SetConnection(msg->m_arg);
				break;
			case EConnManMsg_SetTimeout:
				SetTimeout(msg->m_arg);
				break;
			case EConnManMsg_SetPollInterval:
				SetPollingInterval(msg->m_arg);
				break;
			case EConnManMsg_Kill:
				if ((err = Close()))
					CLog::Error("CConnectionManager::Close() error: 0x%x\n", err);
				return 0;
			default:
				break;
		}
		
		delete msg;
	}
}

//******************************************************************************
//
//******************************************************************************
int CConnectionManager::Initialize()
{
	int err;
	
	// Note that a default connection of 0 has a special meaning... last
	// connection added?? Temp?? Dunno.
	s_DefaultConnection = 1;
	s_DefaultTimeout = 100; // 100 Ticks
	s_DefaultPollingInterval = 200000; // 200 ms
	
	s_Connection = -1;
	s_ConnManState = EConnManState_Disconnected;
	s_ConnectionTimeout = 0;
	
	s_sema = sceKernelCreateSema("ConMan_Error_Flag_Mutex", 0, 1, 1, NULL);
	
	if ((err = pspSdkInetInit()))
	{
		SetError(CON_MAN_ERR_INIT);
		return err;
	}
	
	return 0;
}

//******************************************************************************
//
//******************************************************************************
int CConnectionManager::Close()
{
	sceKernelDeleteMbx(s_mbx);
	
	if (s_ConnManState != EConnManState_Disconnected)
		Disconnect();
	
	sceKernelDeleteSema(s_sema);
	
	pspSdkInetTerm();
	
	return 0;
}

//******************************************************************************
//
//******************************************************************************
int CConnectionManager::Connect(int aConnection)
{
	int err;
	
	s_DisconnectRequested = false; // Abort a potential disconnect request
	ClearError(); // Reset the error flag

	if (sceWlanGetSwitchState() == 0)
	{
		SetError(CON_MAN_ERR_SWITCH_OFF);
		return CON_MAN_ERR_SWITCH_OFF;
	}
	
	if (s_ConnManState == EConnManState_Connected) // Already connected
	{
		if (s_Connection == s_DefaultConnection)
		{
			SetError(CON_MAN_ERR_ALREADY_CONNECTED);
			return CON_MAN_ERR_ALREADY_CONNECTED;
		}
		
		Disconnect();
	}
	else if (s_ConnManState == EConnManState_Connecting) // Connected with a different conn.
	{
		SetError(CON_MAN_ERR_ALREADY_CONNECTING);
		return CON_MAN_ERR_ALREADY_CONNECTING;
	}
	
	s_ConnManState = EConnManState_Connecting;
	s_Connection = aConnection;
	s_ConnectionTimeout = s_DefaultTimeout;
	
	if ((err = sceNetApctlConnect(aConnection)))
	{
		Disconnect();
		SetError(CON_MAN_ERR_CONNECT);
		return err;
	}
	
	int state;
	while (1)
	{
		sceKernelDelayThread(s_DefaultPollingInterval);
		if ((err = sceNetApctlGetState(&state)))
		{
			Disconnect();
			SetError(CON_MAN_ERR_POLLSTATE);
			return err;
		}
		
		// Have we successfully obtained an IP address?
		if (state == 4)
		{
			s_ConnManState = EConnManState_Connected;
			s_ConnectionTimeout = 0;
			return 0;
		}
		else if (state == 0)
		{
			// Usually means we were unable to to contact the wireless AP
			Disconnect();
			SetError(CON_MAN_ERR_AP);
			return 0;
		}
		
		s_ConnectionTimeout--;
		if (s_ConnectionTimeout <= 0)
		{
			Disconnect();
			SetError(CON_MAN_ERR_TIMEOUT);
			return CON_MAN_ERR_TIMEOUT;
		}
	}
	
	return 0;
}

//******************************************************************************
//
//******************************************************************************
int CConnectionManager::Disconnect()
{
	int err;
	
	s_DisconnectRequested = false;
	s_ConnManState = EConnManState_Disconnected;
	s_Connection = -1;
	s_ConnectionTimeout = 0;
	
	if ((err = sceNetApctlDisconnect()))
	{
		SetError(CON_MAN_ERR_DISCONNECT);
		return err;
	}
	
	return 0;
}

//******************************************************************************
//
//******************************************************************************
void CConnectionManager::GetPSPConnections(std::vector<MPSPConnection>& aVector)
{
	if (s_pspConnections.size() == 0)
		CachePSPConnections();
	
	std::vector<MPSPConnection>::iterator it = s_pspConnections.begin();
	while (it != s_pspConnections.end())
	{
		aVector.push_back(*it);
		it++;
	}
}

//******************************************************************************
//
//******************************************************************************
void CConnectionManager::CachePSPConnections()
{
	int i; // Starts at 1, 0 is a special connection
	netData data;
	MPSPConnection conn;
	int netParamRet;
	int consecutiveInvalids = 0;
	
	for (i = 1; (i <= 100 && consecutiveInvalids < 5); i++)
	{
		netParamRet = sceUtilityCheckNetParam(i);
		if (netParamRet == 0)
		{
			// Get the connection name
			sceUtilityGetNetParam(i, 0, &data);
			conn.name = data.asString;
			conn.id = i;
			
			// Is the connection DHCP?
			sceUtilityGetNetParam(i, 4, &data);
			if (data.asUint)
				conn.is_dhcp = false;
			else
				conn.is_dhcp = true;
			
			s_pspConnections.push_back(conn);
			consecutiveInvalids = 0;
		}
		else if ((u32)netParamRet == PSP_NETPARAM_ERROR_BAD_NETCONF)
		{
			consecutiveInvalids += 1;
		}
		else
		{
			// Uhh... corrupt flash??
			CString err;
			err.Printf("Possibly corrupt flash1, run the browser to fix it. Error: 0x%x\n", netParamRet);
			CLog::Error(err);
			ASSERT(0, err.GetPtr());
		}
	}
}

//******************************************************************************
//
//******************************************************************************
CString CConnectionManager::GetPSPConnectionName(u32 aConnId)
{
	int err;
	netData data;
	
	if ((err = sceUtilityCheckNetParam(aConnId)))
	{
		return "";
	}
	
	sceUtilityGetNetParam(aConnId, 0, &data);
	return data.asString;
}

void CConnectionManager::RequestConnect(int aConnection)
{
	int err;
	SConnManMsg* msg = new SConnManMsg;
	
	msg->m_type = EConnManMsg_Connect;
	msg->m_arg = aConnection;
	
	if ((err = sceKernelSendMbx(s_mbx, msg)))
		CLog::Error("RequestConnect => Error sending to connection manager mbx: 0x%x\n", err);
}

void CConnectionManager::RequestDisconnect()
{
	int err;
	SConnManMsg* msg = new SConnManMsg;
	
	msg->m_type = EConnManMsg_Disconnect;
	
	if ((err = sceKernelSendMbx(s_mbx, msg)))
		CLog::Error("RequestDisconnect => Error sending to connection manager mbx: 0x%x\n", err);
}

void CConnectionManager::RequestSetConnection(int aConnection)
{
	int err;
	SConnManMsg* msg = new SConnManMsg;
	
	msg->m_type = EConnManMsg_SetConn;
	msg->m_arg = aConnection;
	
	if ((err = sceKernelSendMbx(s_mbx, msg)))
		CLog::Error("RequestSetConnection => Error sending to connection manager mbx: 0x%x\n", err);
}

void CConnectionManager::RequestSetTimeout(int aTimeout)
{
	int err;
	SConnManMsg* msg = new SConnManMsg;
	
	msg->m_type = EConnManMsg_SetTimeout;
	msg->m_arg = aTimeout;
	
	if ((err = sceKernelSendMbx(s_mbx, msg)))
		CLog::Error("RequestSetTimeout => Error sending to connection manager mbx: %x\n", err);
}

void CConnectionManager::RequestSetPollingInterval(int aInterval)
{
	int err;
	SConnManMsg* msg = new SConnManMsg;
	
	msg->m_type = EConnManMsg_SetPollInterval;
	msg->m_arg = aInterval;
	
	if ((err = sceKernelSendMbx(s_mbx, msg)))
		CLog::Error("RequestSetPollingInterval => Error sending to connection manager mbx: 0x%x\n", err);
}

void CConnectionManager::RequestKill()
{
	int err;
	SConnManMsg* msg = new SConnManMsg;
	
	msg->m_type = EConnManMsg_Kill;
	
	if ((err = sceKernelSendMbx(s_mbx, msg)))
		CLog::Error("RequestKill => Error sending to connection manager mbx: 0x%x\n", err);
}

int CConnectionManager::CheckError()
{
	int toRet;
	sceKernelWaitSema(s_sema, 1, 0);
	toRet = s_ErrFlag;
	sceKernelSignalSema(s_sema, 1);
	return toRet;
}

void CConnectionManager::SetError(int aErr)
{
	sceKernelWaitSema(s_sema, 1, 0);
	s_ErrFlag |= aErr;
	sceKernelSignalSema(s_sema, 1);
}

void CConnectionManager::ClearError()
{
	sceKernelWaitSema(s_sema, 1, 0);
	s_ErrFlag = 0;
	sceKernelSignalSema(s_sema, 1);
}

CString CConnectionManager::GetErrorText(int aErr)
{
	switch (aErr)
	{
		case CON_MAN_ERR_INIT:
			return CString("Error initializing the connection manager.");
		case CON_MAN_ERR_TERM:
			return CString("Error terminating the connection manager.");
		case CON_MAN_ERR_CONNECT:
			return CString("Error bringing up the connection.");
		case CON_MAN_ERR_ALREADY_CONNECTED:
			return CString("Currently connected.");
		case CON_MAN_ERR_ALREADY_CONNECTING:
			return CString("Currently establishing a connection.");
		case CON_MAN_ERR_DISCONNECT:
			return CString("Error tearing down the connection.");
		case CON_MAN_ERR_POLLSTATE:
			return CString("Error polling the connection state.");
		case CON_MAN_ERR_TIMEOUT:
			return CString("Connection timed out.");
		case CON_MAN_ERR_SWITCH_OFF:
			return CString("WLAN switch is off.");
		case CON_MAN_ERR_AP:
			return CString("Error contacting the wireless access point.");
		default:
			CString toRet;
			toRet.Printf("Multiple connection manager errors: 0x%x.", aErr);
			return toRet;
	}
}
