/*******************************************************************************

  PSPoste - Wireless Email Client for the Sony PSP
  Copyright (C) Benjamin Biron, Sajeeth Cherian, David Cummings

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

  Module :	CConnectionManager.h

  Description :	

  Last Modified $Date: 2008/08/31 03:50:11 $

  $Revision: 1.4 $

*******************************************************************************/

#ifndef CCONNECTIONMANAGER_H_
#define CCONNECTIONMANAGER_H_

//******************************************************************************
//   Include Files
//******************************************************************************
#include <stdlib.h>
#include <vector>
#include <pspthreadman.h>
#include "CString.h"

//******************************************************************************
//   Macros
//******************************************************************************

//******************************************************************************
//   Types
//******************************************************************************
#define CON_MAN_ERR_INIT               (1 << 0) // Calling nlhInit failed
#define CON_MAN_ERR_TERM               (1 << 1) // Calling nlhTerm failed
#define CON_MAN_ERR_CONNECT            (1 << 2) // Calling sceNetApctlConnect failed
#define CON_MAN_ERR_ALREADY_CONNECTED  (1 << 3) // Called Connect() when connection already up
#define CON_MAN_ERR_ALREADY_CONNECTING (1 << 4) // Called Connect() when already connecting
#define CON_MAN_ERR_DISCONNECT         (1 << 5) // Calling sceNetApctlDisconnect failed
#define CON_MAN_ERR_POLLSTATE          (1 << 6) // Calling sceNetApctlGetState failed
#define CON_MAN_ERR_TIMEOUT            (1 << 7) // A timeout has occurred
#define CON_MAN_ERR_SWITCH_OFF         (1 << 8) // WLAN switch is off
#define CON_MAN_ERR_AP                 (1 << 9) // Error connecting to the wireless AP

typedef enum _EConnManState
{
	EConnManState_Connected,
	EConnManState_Connecting,
	EConnManState_Disconnected,
	
	EConnManState_MAX
} EConnManState;

typedef enum _EConnManMsgType
{
	EConnManMsg_Connect,
	EConnManMsg_Disconnect,
	EConnManMsg_SetConn,
	EConnManMsg_SetTimeout,
	EConnManMsg_SetPollInterval,
	EConnManMsg_Kill,
	
	EConnManMsg_MAX
} EConnManMsgType;

typedef struct _SConnManMsg
{
	struct _SConnManMsg* link;
	EConnManMsgType m_type;
	int m_arg;
} SConnManMsg;

typedef struct _MPSPConnection
{
	CString name;
	u32 id;
	bool is_dhcp;
} MPSPConnection;

//******************************************************************************
//   Constants
//******************************************************************************

//******************************************************************************
//   Class definitions
//******************************************************************************

//******************************************************************************
//   CConnectionManager: The functions marked as "NOT THREAD SAFE" should not be
//	 called by user code! Use the mailbox instead.
//
//	 Note that some state vars are accessed by seperate threads and left
//	 unprotected. This is because they are written only by 1 thread.
//******************************************************************************
class CConnectionManager
{
	public: // Thread callback function
	
		static int ConnManThreadCB(SceSize args, void* argp);
	
	public: // New methods
	
		// Is a connection currently up?
		inline static EConnManState GetState() {return s_ConnManState;}
		
		// Which connection is up? -1 means not connected
		inline static int GetConnection() {return s_Connection;}
		
		// Get/Clear the value of the error flag
		static int CheckError();
		static void ClearError();
		
		// Return the textual value of the error flag
		static CString GetErrorText(int aErr);
		
		// Copies the list of cached PSP connections to aVector
		static void GetPSPConnections(std::vector<MPSPConnection>& aVector);
		
		// Gets the name of a PSP connection
		static CString GetPSPConnectionName(u32 aConnId);
		
	public: // These methods form a msg and send it to the conn man thread
	
		static void RequestConnect(int aConnection = s_DefaultConnection);
		static void RequestDisconnect();
		static void RequestSetConnection(int aConnection);
		static void RequestSetTimeout(int aTimeout);
		static void RequestSetPollingInterval(int aInterval);
		static void RequestKill();
	
	private: // New methods
		
		// Initialize the connection manager
		static int Initialize();
		
		// Brings down the connection if it's up
		static int Close();
		
		// Bring up the specified connection - NOT THREAD SAFE!
		static int Connect(int aConnection = s_DefaultConnection);
		
		// Bring down the network connection - NOT THREAD SAFE!
		static int Disconnect();
		
		// Set the bits in the error flag
		static void SetError(int aErr);
		
		// Sets the connection setting to use when left unspecified  - NOT THREAD SAFE!
		inline static void SetConnection(int aConnection)
		                   {s_DefaultConnection = aConnection;}
		
		// Set the timeout in ticks - NOT THREAD SAFE!
		inline static void SetTimeout(int aTimeout)
		                   {s_DefaultTimeout = aTimeout;}
		
		// Sets the polling granularity (after how many frames should the
		// connection status be polled) - NOT THREAD SAFE!
		inline static void SetPollingInterval(int aInterval)
		                   {s_DefaultPollingInterval = aInterval;}

		// Caches the list of PSP connections
		static void CachePSPConnections();
		
	private: // Data
		
		// These variables are modified by setters
		static int s_DefaultConnection;
		static unsigned int s_DefaultTimeout;
		static unsigned int s_DefaultPollingInterval;
		
		// These variables are status variables
		static int s_Connection;
		static EConnManState s_ConnManState;
		static int s_ConnectionTimeout;
		static int s_ErrFlag;
		static bool s_DisconnectRequested;
		
		// The mbx used to send commands to the connection manager thread
		static SceUID s_mbx;
		
		// The mutex for the error flag
		static SceUID s_sema;
		
		// This is the cached network connections. It's cached since we can't
		// stop iterating the first time sceUtilityCheckNetParam fails since
		// there's no guarantee that the network connections will be contiguous.
		// So to avoid looping over 100 possible network connections everytime,
		// we just do it the first time then copy this result in vectors passed
		// in to GetPSPConnections.
		static std::vector<MPSPConnection> s_pspConnections;

};

#endif
