/*******************************************************************************

  PSPoste - Wireless Email Client for the Sony PSP
  Copyright (C) Benjamin Biron, Sajeeth Cherian, David Cummings

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

  Module :	CDirectoryList.cpp

  Description :	

  Last Modified $Date: 2007/01/01 21:55:47 $

  $Revision: 1.4 $

*******************************************************************************/

//******************************************************************************
//   Include Files
//******************************************************************************
#include "CContextMenu.h"
#include "CTextureManager.h"
#include "CFont.h"
#include "CInput.h"
#include "CGfx.h"
#include "CSkinManager.h"
#include "CContextMenuItem.h"
#include "CProcess.h"
#include "CFrameWork.h"
#include "CHUHelp.h"
#include "irda_KeyboardLayouts.h"
#include "CKeyboardNavWrapper.h"

//******************************************************************************
//   Local Macros
//******************************************************************************

//******************************************************************************
//   Local Constants
//******************************************************************************
static const float WINDOW_BORDER(1.f); // Border around context menu (all sides)
static const float WINDOW_PADDING(2.f); // Padding between border and menu item (all sides)
static const float WINDOW_ITEM_PADDING(2.f); // Vert padding between menu items
static const float WINDOW_MINIMUM_WIDTH(75.f); // actually the min width of item
static const float WINDOW_DIVIDER_HEIGHT(1.f);
static const float WINDOW_DIVIDER_PADDING(1.f); // top and bottom
static const float WINDOW_SUB_MENU_ADJUST(10.f);

static const int SUB_MENU_RETURNED_NO_SELECTION = -2;
//******************************************************************************
//   Static Prototypes
//******************************************************************************

//******************************************************************************
//   Global Variables
//******************************************************************************

//******************************************************************************
//   Static Variables
//******************************************************************************

//******************************************************************************
//   Class Definition
//******************************************************************************

//******************************************************************************
//	
//******************************************************************************
CContextMenu::CContextMenu(CContextMenuItemList * const menuItemList, V2 const menuPosition, bool root) :
	m_pSubMenu(NULL),
	m_pMenuItemList(menuItemList),
	m_menuSize(0.f, 0.f),
	m_menuItemSize(WINDOW_MINIMUM_WIDTH, 0.f),
	m_menuPosition(menuPosition),
	m_exitValue(-1),
	m_dividers(0),
	m_root(root)
{
	// Set up HUD
	CHUHelp::Clear();
	CHUHelp::SetButton(CHUHelp::BUTTON_UP, "Next Item");
	CHUHelp::SetButton(CHUHelp::BUTTON_DOWN, "Previous Item");
	CHUHelp::SetButton(CHUHelp::BUTTON_LEFT, "Expand sub menu");
	CHUHelp::SetButton(CHUHelp::BUTTON_CROSS, "Select Item");

	// Iterate through all of the items and determine the biggest menu item size
	CContextMenuItemList::iterator it = m_pMenuItemList->begin();

	m_SelectedItemIndex = 0; // Select the first object

	// Set menuItemSize to largest width/height of items, count # of dividers
	while( it !=  m_pMenuItemList->end() ) {
		V2 text_size( CFont::GetActiveUiFont()->GetStringSize( (*it)->m_szName) );
		m_menuItemSize.x = (text_size.x > m_menuItemSize.x) ? text_size.x : m_menuItemSize.x;
		m_menuItemSize.y = (text_size.y > m_menuItemSize.y) ? text_size.y : m_menuItemSize.y;
		
		if ( (*it)->m_hasDivider ) {
			m_dividers++;
		}
		it++;
	}

	// Derive the menuSize based on the # of items, their size and the window properties
	m_menuSize.x = m_menuItemSize.x + 
					(WINDOW_BORDER * 2) +
					(WINDOW_PADDING * 2);

	m_menuSize.y = (m_pMenuItemList->size() * m_menuItemSize.y) + 
					(m_pMenuItemList->size() * WINDOW_ITEM_PADDING) + 
					(m_dividers * (WINDOW_DIVIDER_HEIGHT + 2 * WINDOW_DIVIDER_PADDING)) +
					(WINDOW_BORDER * 2) +
					(WINDOW_PADDING * 2);

	SetFocus( true );
}

//******************************************************************************
//	
//******************************************************************************
CContextMenu::~CContextMenu()
{
	if (m_root) {
		SAFE_DELETE(m_pSubMenu);

		// iterate through the list of menu items and delete each one
		CContextMenuItemList::iterator it = m_pMenuItemList->begin();
		while( it !=  m_pMenuItemList->end() ) {
			SAFE_DELETE((*it));
			it++;
		}

		// delete the list	
		SAFE_DELETE(m_pMenuItemList);
	}
}	

//******************************************************************************
//	
//******************************************************************************
void	CContextMenu::Render()
{
	// Draw a drop shadow, gotta figure out how to blur this puppy
	CGfx::DrawQuad( V2( CGfx::s_ScreenWidth - m_menuSize.x - m_menuPosition.x + 1, 0 + m_menuPosition.y + 1 ),
				   V2( m_menuSize.x, m_menuSize.y),
				   CSkinManager::GetColor( "context_menu", "shadow_color", 0xFF00FF00));

	// Draw the borders
    // Sort of cheating here, couldn't send bgcolor to none opaque, if we want that functionality
	// need to draw 4 seperate quads
	CGfx::DrawQuad( V2( CGfx::s_ScreenWidth - m_menuSize.x - m_menuPosition.x, 0 + m_menuPosition.y ),
					V2( m_menuSize.x, m_menuSize.y),
					CSkinManager::GetColor( "context_menu", "border_color", 0xFF00FF00));

	// Draw menu background (overlay over the menu border quad)
	CGfx::DrawQuad( V2( CGfx::s_ScreenWidth - m_menuSize.x + WINDOW_BORDER - m_menuPosition.x, 0 + WINDOW_BORDER + m_menuPosition.y),
					V2( m_menuSize.x - WINDOW_BORDER * 2, m_menuSize.y - WINDOW_BORDER * 2),
					CSkinManager::GetColor( "context_menu", "background_color", 0xFF000000));

	// Initialize the text position coordinates
	float textPositionX = CGfx::s_ScreenWidth - m_menuSize.x + WINDOW_BORDER + WINDOW_PADDING - m_menuPosition.x;
	float textPositionY = WINDOW_BORDER + WINDOW_PADDING + m_menuPosition.y;  

	// Print the text for each menu item
	CContextMenuItemList::iterator it = m_pMenuItemList->begin();
	unsigned int itemIndex = 0; // Used to determine if item is selected (could do .at(selectedItemIndex on iterator) but this is process intensive
    while( it !=  m_pMenuItemList->end() ) {
		textPositionY += WINDOW_ITEM_PADDING / 2;
		V2 text_size( CFont::GetActiveUiFont()->GetStringSize( (*it)->m_szName) );
        V2 text_position( textPositionX, textPositionY + m_menuItemSize.y - text_size.y );
		
		// Print regular text for unselected, print selected text and box for selected item
		if (itemIndex != m_SelectedItemIndex) {
			CFont::GetActiveUiFont()->Print( (*it)->m_szName, text_position, CSkinManager::GetColor( "context_menu", "text_color", 0xFFFFFFFF));
		} else {
			if (m_pSubMenu == NULL) {
				CGfx::DrawQuad( V2( textPositionX - WINDOW_PADDING, textPositionY - WINDOW_ITEM_PADDING / 2 ),
						V2( m_menuItemSize.x + WINDOW_PADDING * 2, m_menuItemSize.y + WINDOW_ITEM_PADDING ),
						CSkinManager::GetColor( "context_menu", "selected_item_hilight_color", 0xFFFFFFFF));
			} else {
				CGfx::DrawQuad( V2( textPositionX - WINDOW_PADDING, textPositionY - WINDOW_ITEM_PADDING / 2 ),
						V2( m_menuItemSize.x + WINDOW_PADDING * 2, m_menuItemSize.y + WINDOW_ITEM_PADDING ),
						CSkinManager::GetColor( "context_menu", "selected_item_hilight_half_color", 0xFFFFFFFF));
			}
			CFont::GetActiveUiFont()->Print( (*it)->m_szName, text_position, CSkinManager::GetColor( "context_menu", "selected_text_color", 0xFF000000));
		}	

		// Move text position to next item
		textPositionY += m_menuItemSize.y + WINDOW_ITEM_PADDING / 2;

		// If this item has divider, print divider above next item
		if ((*it)->m_hasDivider) {
			textPositionY += WINDOW_DIVIDER_PADDING;
			CGfx::DrawQuad( V2( textPositionX - WINDOW_PADDING, textPositionY),
                    V2( m_menuItemSize.x + WINDOW_PADDING * 2, WINDOW_DIVIDER_HEIGHT ),
                    CSkinManager::GetColor( "context_menu", "border_color", 0xFF00FF00));
			textPositionY += WINDOW_DIVIDER_HEIGHT + WINDOW_DIVIDER_PADDING;
		}
	
		
        it++;
		itemIndex++;
    }
}

//******************************************************************************
//	
//******************************************************************************
void	CContextMenu::Process()
{
	if (m_pSubMenu != NULL) {
		if (m_pSubMenu->getExitValue() != -1) {
			if (m_pSubMenu->getExitValue() == SUB_MENU_RETURNED_NO_SELECTION) {
				SAFE_DELETE(m_pSubMenu);
			} else {
				m_exitValue = m_pSubMenu->getExitValue();
			}
		}
	}
	CWindow::Process();
}

//******************************************************************************
//
//******************************************************************************
bool	CContextMenu::ProcessInput()
{
	if (m_pSubMenu == NULL) {
		if (CKeyboardNavWrapper::ActivateControl()) {
			if (m_pMenuItemList->at(m_SelectedItemIndex)->m_pSubItemList == NULL) {
				m_exitValue = m_pMenuItemList->at(m_SelectedItemIndex)->m_action;
			} else {
				buildSubMenu();
			}
		} 

		if (CKeyboardNavWrapper::Up()) {
			// Move to the previous item in the list
			if ( m_SelectedItemIndex != 0 ) { // can't decrement past first item
				m_SelectedItemIndex--;
			} else {
				m_SelectedItemIndex = m_pMenuItemList->size() -1 ;
			}
		}

		if (CKeyboardNavWrapper::Down()) {
			// Move to the next item in the list
			if ( m_SelectedItemIndex != m_pMenuItemList->size() - 1) { // can't increment past last item
				m_SelectedItemIndex++;
			} else {
				m_SelectedItemIndex = 0; //wrap to the top (your welcome ben)
			}
		}
	
		if (CKeyboardNavWrapper::CollapseSubContextMenu()) {
			if (!m_root) {
				m_exitValue = SUB_MENU_RETURNED_NO_SELECTION;
			}
		}

		if (CKeyboardNavWrapper::ExpandSubContextMenu()) {
			if (m_pMenuItemList->at(m_SelectedItemIndex)->m_pSubItemList != NULL) {			
				buildSubMenu();
			}
		}
	}
	return false;
}

//******************************************************************************
//
//******************************************************************************
int	CContextMenu::getExitValue() {
	return m_exitValue;
}

//******************************************************************************
//
//******************************************************************************
void CContextMenu::buildSubMenu() {
	V2 subMenuPosition(0, 0);

	float subMenuItemWidth = WINDOW_MINIMUM_WIDTH;

	std::vector< CContextMenuItem * >::iterator it = m_pMenuItemList->at(m_SelectedItemIndex)->m_pSubItemList->begin();
	// Set menuItemSize to largest width/height of items, count # of dividers
	while( it !=  m_pMenuItemList->at(m_SelectedItemIndex)->m_pSubItemList->end() ) {
		V2 text_size( CFont::GetActiveUiFont()->GetStringSize( (*it)->m_szName) );
		subMenuItemWidth = (text_size.x > m_menuItemSize.x) ? text_size.x : m_menuItemSize.x;
		it++;
	}
	subMenuPosition.x = m_menuPosition.x + m_menuSize.x - WINDOW_SUB_MENU_ADJUST;
	float textPositionY = WINDOW_BORDER + WINDOW_PADDING + m_menuPosition.y;  
	it = m_pMenuItemList->at(m_SelectedItemIndex)->m_pSubItemList->begin();
	unsigned int itemIndex = 0; // Used to determine if item is selected (could do .at(selectedItemIndex on iterator) but this is process intensive
    while( it !=  m_pMenuItemList->at(m_SelectedItemIndex)->m_pSubItemList->end() ) {
		textPositionY += WINDOW_ITEM_PADDING / 2;
		V2 text_size( CFont::GetActiveUiFont()->GetStringSize( (*it)->m_szName) );
        V2 text_position( 0, textPositionY + m_menuItemSize.y - text_size.y );
		// Print regular text for unselected, print selected text and box for selected item
		if (itemIndex == m_SelectedItemIndex) {
			break;
		}	
		// Move text position to next item
		textPositionY += m_menuItemSize.y + WINDOW_ITEM_PADDING / 2;
		// If this item has divider, print divider above next item
		if ((*it)->m_hasDivider) {
			textPositionY += WINDOW_DIVIDER_PADDING;
			textPositionY += WINDOW_DIVIDER_HEIGHT + WINDOW_DIVIDER_PADDING;
		}
        it++;
		itemIndex++;
    }
	subMenuPosition.y = textPositionY + m_menuItemSize.y/2;// + m_pMenuItemList->size() * m_menuItemSize.y - m_menuItemSize.y/2;

	if (m_pMenuItemList->at(m_SelectedItemIndex)->m_pSubItemList != NULL) {
		m_pSubMenu = new CContextMenu(m_pMenuItemList->at(m_SelectedItemIndex)->m_pSubItemList,
									  subMenuPosition,
									  false);
	}
}

//*******************************  END OF FILE  ********************************
