/*******************************************************************************

  PSPoste - Wireless Email Client for the Sony PSP
  Copyright (C) Benjamin Biron, Sajeeth Cherian, David Cummings

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

  Module :	CControlSelect.cpp

  Description :	

  Last Modified $Date: 2007/01/27 18:05:50 $

  $Revision: 1.5 $

  Copyright (C) David Cummings, Sajeeth Cherian and Ben Biron

*******************************************************************************/

//******************************************************************************
//   Include Files
//******************************************************************************
#include "CControlSelect.h"
#include "CTextureManager.h"
#include "CFont.h"
#include "CInput.h"
#include "CGfx.h"
#include "CSkinManager.h"
#include "CFrameWork.h"
#include "CControlSelect.h"
#include "CMath.h"
#include "CHUHelp.h"
#include "irda_KeyboardLayouts.h"
#include "CKeyboardNavWrapper.h"

//******************************************************************************
//   Local Macros
//******************************************************************************

//******************************************************************************
//   Local Constants
//******************************************************************************
static const float	WINDOW_BORDER( 1.f ); // Border around context menu (all sides)
static const float  WINDOW_PADDING( 1.f ); // Padding between border and menu item (all sides)
static const float  WINDOW_ITEM_PADDING ( 2.f ); // Vert padding between menu items

//******************************************************************************
//   Static Prototypes
//******************************************************************************

//******************************************************************************
//   Global Variables
//******************************************************************************

//******************************************************************************
//   Static Variables
//******************************************************************************

//******************************************************************************
//   Class Definition
//******************************************************************************

//******************************************************************************
//	
//******************************************************************************
CControlSelect::CControlSelect(V2 const size, V2 const position, int maxDisplayedItems, CControlSelectItemList* const itemList, CControlSelectItem* selectedItem) :
	m_size(size),
	m_position(position),
	m_maxDisplayedItems(maxDisplayedItems),
	m_pItemList(itemList),
	m_pDisplayedItemList(NULL),
	m_pSelectedItem(selectedItem),
	m_pDisplayedItem(NULL),
	m_pHistoricItem(NULL)
{
	initList(itemList, selectedItem);
}

//******************************************************************************
//	
//******************************************************************************
CControlSelect::~CControlSelect()
{
	destroyList();
}	

//******************************************************************************
//	
//******************************************************************************
void CControlSelect::initList(CControlSelectItemList* const itemList, CControlSelectItem* selectedItem)
{
	m_pItemList = itemList;
	m_pDisplayedItemList = NULL;
	m_pSelectedItem = selectedItem;
	m_pDisplayedItem = NULL;
	m_pHistoricItem = NULL;

	m_pDisplayedItemList = new CControlSelectItemList();

	// If no item is specified set selected item to first item
	if (m_pSelectedItem == NULL) {
		m_pSelectedItem = *(m_pItemList->begin());
		m_pDisplayedItem = m_pSelectedItem;
		buildDisplayList();
	} else {
		m_pDisplayedItem = *(m_pItemList->begin());
		buildDisplayList();
		updateFirstDisplayed();
	}	
}

//******************************************************************************
//	
//******************************************************************************
void CControlSelect::destroyList()
{
	// Delete of the items in the list
	CControlSelectItemList::iterator it = m_pItemList->begin();
	while( it !=  m_pItemList->end() ) {
		SAFE_DELETE(*it);
		it++;
	}

	// Remove the displayed item list
	SAFE_DELETE(m_pDisplayedItemList);

	// Remove the whole list
	SAFE_DELETE(m_pItemList);
}

//******************************************************************************
//	
//******************************************************************************
void CControlSelect::rebuildList(CControlSelectItemList* const itemList, CControlSelectItem* selectedItem)
{
	destroyList();
	initList(itemList, selectedItem);
}


//******************************************************************************
//	
//******************************************************************************
void	CControlSelect::RenderControl()
{
	if(isHidden()) {
		return;
	}

	ARGB itemColor, backgroundColor;

	// Draw the borders
    // Sort of cheating here, couldn't send bgcolor to none opaque, if we want that functionality
	// need to draw 4 seperate quads
	if (isSelected())
		itemColor = CSkinManager::GetColor( "select_box", "selected_border_color", 0xFFAAFFAA);
	else
		itemColor = CSkinManager::GetColor( "select_box", "border_color", 0xFFAAFFAA);
		
	CGfx::DrawQuad( V2( m_position.x, m_position.y ),
					V2( m_size.x, m_size.y),
					itemColor);
	
	// Draw menu background (overlay over the menu border quad)
	if (isEnabled())
		backgroundColor = CSkinManager::GetColor( "select_box", "background_color", 0xFFAAFFAA);
	else
		backgroundColor = CSkinManager::GetColor( "select_box", "disabled_background_color", 0xFFAAFFAA);
		
	CGfx::DrawQuad( V2( m_position.x + WINDOW_BORDER, m_position.y + WINDOW_BORDER),
					V2( m_size.x - WINDOW_BORDER * 2, m_size.y - WINDOW_BORDER * 2),
					backgroundColor);

	// Draw select box arrow box and arrow
	CGfx::DrawQuad( V2( m_position.x + m_size.x - m_size.y, m_position.y),
					V2( m_size.y, m_size.y),
					itemColor);
	// Draw Arrow
	float baseWidth = (m_size.y - 2 * WINDOW_BORDER - 2 * WINDOW_PADDING) / 2;
	float triangleHeight = SQRTF( SQUARE(baseWidth) + SQUARE(baseWidth/2));
	sVertexColor *	p_vert;
	CGfx::GetPolyList( 3, &p_vert );
	p_vert[ 0 ].color = CSkinManager::GetColor( "select_box", "background_color", 0xFF00FF00);
	p_vert[ 0 ].pos.x = m_position.x + m_size.x - m_size.y /2 - baseWidth/2;
	p_vert[ 0 ].pos.y = m_position.y + m_size.y / 2 - triangleHeight/2;
	p_vert[ 0 ].pos.z = 0.f;
	p_vert[ 1 ].color = CSkinManager::GetColor( "select_box", "background_color", 0xFF00FF00);
	p_vert[ 1 ].pos.x = p_vert[ 0 ].pos.x + baseWidth;
	p_vert[ 1 ].pos.y = p_vert[ 0 ].pos.y;
	p_vert[ 1 ].pos.z = 0.f;
	p_vert[ 2 ].color = CSkinManager::GetColor( "select_box", "background_color", 0xFF00FF00);
	p_vert[ 2 ].pos.x = m_position.x + m_size.x - m_size.y/2;
	p_vert[ 2 ].pos.y = p_vert[ 0 ].pos.y + triangleHeight;
	p_vert[ 2 ].pos.z = 0.f;
	CGfx::DrawPoly2D( GU_TRIANGLE_STRIP, p_vert, 3 );

	// Draw text
	V2 text_size( CFont::GetActiveUiFont()->GetStringSize( m_pSelectedItem->m_text ));
    V2 text_position( m_position.x + WINDOW_BORDER + WINDOW_PADDING, m_position.y + m_size.y - WINDOW_BORDER - WINDOW_PADDING - text_size.y );
	CFont::GetActiveUiFont()->Print( m_pSelectedItem->m_text, text_position, CSkinManager::GetColor( "select_box", "text_color", 0xFFFFFFFF));

	// Draw drop down window
	if (isActive()) {
		// Set up HUD
		CHUHelp::Clear();
		CHUHelp::SetButton(CHUHelp::BUTTON_UP, "Previous Item");
		CHUHelp::SetButton(CHUHelp::BUTTON_DOWN, "Next Item");
		CHUHelp::SetButton(CHUHelp::BUTTON_TRIANGLE, "Cancel");
		CHUHelp::SetButton(CHUHelp::BUTTON_CROSS, "Select Item");

		V2 menuItemSize(0, 0);
		V2 menuSize(0, 0);
		V2 startPosition(0, 0);

		// Iterate through all of the items and determine the biggest menu item size
		CControlSelectItemList::iterator it = m_pDisplayedItemList->begin();

		// Set menuItemSize to largest width/height of items
		while( it !=  m_pDisplayedItemList->end() ) {
			V2 text_size( CFont::GetActiveUiFont()->GetStringSize( (*it)->m_text ));
			menuItemSize.x = (text_size.x > menuItemSize.x) ? text_size.x : menuItemSize.x;
			menuItemSize.y = (text_size.y > menuItemSize.y) ? text_size.y : menuItemSize.y;
			it++;
		}

		// Derive the menuSize based on the # of items, their size and the window properties
		menuSize.x = m_size.x;

		menuSize.y = (m_pDisplayedItemList->size() * menuItemSize.y) + 
						(m_pDisplayedItemList->size() * WINDOW_ITEM_PADDING) + 
						(WINDOW_BORDER * 2) +
						(WINDOW_PADDING * 2);
	
		startPosition.x = m_position.x;
		startPosition.y = m_position.y + m_size.y - 1 ;

		// Draw the borders
		// Sort of cheating here, couldn't send bgcolor to none opaque, if we want that functionality
		// need to draw 4 seperate quads
		CGfx::DrawQuad( startPosition,
						menuSize,
						itemColor);


		// Draw menu background (overlay over the menu border quad)
		CGfx::DrawQuad( V2( startPosition.x + WINDOW_BORDER, startPosition.y + WINDOW_BORDER),
						V2( menuSize.x - WINDOW_BORDER * 2, menuSize.y - WINDOW_BORDER * 2),
						CSkinManager::GetColor( "select_box", "list_background_color", 0xFF000000));

		// Initialize the text position coordinates
		float textPositionX = startPosition.x + WINDOW_BORDER + WINDOW_PADDING;
		float textPositionY = startPosition.y + WINDOW_BORDER + WINDOW_PADDING;  

		// Print the text for each menu item
		it = m_pDisplayedItemList->begin();
		while( it !=  m_pDisplayedItemList->end() ) {
			textPositionY += WINDOW_ITEM_PADDING / 2;
			V2 text_size( CFont::GetActiveUiFont()->GetStringSize( (*it)->m_text ));
			V2 text_position( textPositionX, textPositionY + menuItemSize.y - text_size.y );
			
			// Print regular text for unselected, print selected text and box for selected item
			if ((*it) != m_pSelectedItem) {
				CFont::GetActiveUiFont()->Print( (*it)->m_text, text_position, CSkinManager::GetColor( "select_box", "text_color", 0xFFFFFFFF));
			} else {
				CGfx::DrawQuad( V2( textPositionX - WINDOW_PADDING, textPositionY - WINDOW_ITEM_PADDING / 2 ),
						V2( m_size.x - WINDOW_BORDER * 2, menuItemSize.y + WINDOW_ITEM_PADDING ),
						CSkinManager::GetColor( "select_box", "selected_item_hilight_color", 0xFFFFFFFF));
				CFont::GetActiveUiFont()->Print( (*it)->m_text, text_position, CSkinManager::GetColor( "select_box", "selected_text_color", 0xFF000000));
			}	

			// Move text position to next item
			textPositionY += menuItemSize.y + WINDOW_ITEM_PADDING / 2;	
			it++;
		}
	}
}

//******************************************************************************
//
//******************************************************************************
void	CControlSelect::ProcessControlInput()
{
	CControlSelectItemList::iterator it;

	//Process select box control input
	if (isActive()) {
		// Shift selected item up one unless already first in list
		if (CKeyboardNavWrapper::Up()) {
			it = m_pItemList->begin();
			if (m_pSelectedItem != (*it)) {
					while( it !=  m_pItemList->end() ) {
						if (m_pSelectedItem == (*it)) {
							it--;
							m_pSelectedItem = (*it);
							updateFirstDisplayed();
							return;
						}
						it++;
					}
			}
			//new code dec5, gotta test this
			else {
				m_pSelectedItem = *(--m_pItemList->end());
				updateFirstDisplayed();
				return;
			}
		}

		// Shift selected item down one unless already last in list
		if (CKeyboardNavWrapper::Down()) {
			it = m_pItemList->begin();
			while( it !=  m_pItemList->end() ) {
				if (m_pSelectedItem == (*it)) {
					if (++it != m_pItemList->end()) {
						m_pSelectedItem = (*it);
						updateFirstDisplayed();
						return;
					}
					//new code dec 5, gotta test this!!!
					else {
					    m_pSelectedItem = *(m_pItemList->begin());
						updateFirstDisplayed();
						return;
					}
					// hmmm do i need this anymore???
					continue; //already incremented
				}
				it++;
			}
		}

		if (CKeyboardNavWrapper::SaveComboBoxSelection()) {
			if (m_pSelectedItem != m_pHistoricItem) {
				notifyEvent(CHANGE);
			}
			deActivate();		
		}
		
		if (CKeyboardNavWrapper::CancelComboBoxSelection()) {
			m_pSelectedItem = m_pHistoricItem; // revert
			m_pDisplayedItem = m_pSelectedItem;
			buildDisplayList();			
			deActivate();
		} 
	}
}

//******************************************************************************
// This method assumes that the m_pDisplayedList is properly set, and has 
// enough items underneath to properly display a full select menu
//******************************************************************************
void	CControlSelect::buildDisplayList()
{
	// Remove existing items from the list
	m_pDisplayedItemList->clear();

	// Build the list of items to be displayed, starting with the item in m_pDisplayedItem
	CControlSelectItemList::iterator it = m_pItemList->begin();
	while( it !=  m_pItemList->end() ) {
		// Found the displayed item, display the next items
		if (m_pDisplayedItem == (*it)) {
			//Ensure doesn't reach end (this will happen if list.size() < max items
			for(int x=0; x<m_maxDisplayedItems && it != m_pItemList->end(); x++) {
				m_pDisplayedItemList->push_back(*it);
				it++;
			}
			return;
		}
		it++;
	}
}

//******************************************************************************
//
//******************************************************************************
void	CControlSelect::updateFirstDisplayed() {
	
	CControlSelectItemList::iterator it;

	// Don't change if selected item is already in the list
	it = m_pDisplayedItemList->begin();
	while( it !=  m_pDisplayedItemList->end() ) {
		if (m_pSelectedItem == (*it)) {
			return;
		}
		it++;
	}

	// If selected item found before first display, first display set to selected item and rebuild list
	it = m_pItemList->begin();
	while( it !=  m_pItemList->end() ) {
		if ((*it) == m_pDisplayedItem) break; // selected item must be after list
		if ((*it) == m_pSelectedItem) {
			m_pDisplayedItem = m_pSelectedItem;
			buildDisplayList();
			return;	
		}
		it++;
	}

	// selected must be after
	it = m_pItemList->begin();
	while( it !=  m_pItemList->end() ) {
		if ((*it) == m_pDisplayedItem) {
			m_pDisplayedItem = *(++it); // Move displayed up by one
			buildDisplayList();
			updateFirstDisplayed(); // re run to make sure actually in list!
			return;
		}
		it++;
	}
}

//******************************************************************************
//
//******************************************************************************
CString CControlSelect::getSelectedText() {
	return m_pSelectedItem->m_text;
}

//******************************************************************************
//
//******************************************************************************
u32 CControlSelect::getSelectedValue() {
	return m_pSelectedItem->m_value;
}

//******************************************************************************
// 
//******************************************************************************
void CControlSelect::ProcessControl() {

}

//******************************************************************************
// 
//******************************************************************************
void CControlSelect::activate() {
	CControl::activate();
	m_pHistoricItem = m_pSelectedItem;
}

void CControlSelect::subscribe(MControlSelectEventListener* listener)
{
	m_listeners.push_back(listener);
}

void CControlSelect::notifyEvent(eEvents aEvent)
{
	std::list<MControlSelectEventListener*>::iterator it = m_listeners.begin();

	while (it != m_listeners.end())
	{
		switch (aEvent)
		{
			case CHANGE:
				(*it)->onChange(this);
				break;
			default:
				break;
		}
		it++;
	}
}
//*******************************  END OF FILE  ********************************
