/*******************************************************************************

  PSPoste - Wireless Email Client for the Sony PSP
  Copyright (C) Benjamin Biron

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

  Module :	CDialogModal.cpp

  Description :	

*******************************************************************************/
#include "CDialogModal.h"
#include "CRenderable.h"
#include "CControlButton.h"
#include "CControlManager.h"
#include "CSkinManager.h"
#include "CFont.h"

TModalList CDialogModal::s_modalList;

static const float WINDOW_BORDER(2.f); // Border around context menu (all sides)
static const float WINDOW_PADDING(8.f); // Padding between border and menu item (all sides)
static const float ITEM_PADDING(4.f);
static const float BUTTON_HEIGHT(20.f);
static const float BUTTON_WIDTH(85.f);
static const float MIN_WIDTH_ONE_BUTTON(WINDOW_BORDER*2 + WINDOW_PADDING*2 + BUTTON_WIDTH);
static const float MIN_WIDTH_TWO_BUTTONS(WINDOW_BORDER*2 + WINDOW_PADDING*2 + BUTTON_WIDTH*2 + ITEM_PADDING);
static const float MIN_WIDTH_THREE_BUTTONS(WINDOW_BORDER*2 + WINDOW_PADDING*2 + BUTTON_WIDTH*3 + ITEM_PADDING*2);

static const char* s_buttonCaptions[DIALOG_CONFIG_MAX][MAX_MODAL_DLG_BUTTONS] =
{
	{"OK",     "",            ""},
	{"OK",     "CANCEL",      ""},
	{"YES",    "NO",          ""},
	{"ACCEPT", "ACCEPT ONCE", "DECLINE"}
};

static EDialogButtonValue s_buttonValues[DIALOG_CONFIG_MAX][MAX_MODAL_DLG_BUTTONS] =
{
	{DIALOG_BUTTON_OK,     DIALOG_BUTTON_INVALID,     DIALOG_BUTTON_INVALID},
	{DIALOG_BUTTON_OK,     DIALOG_BUTTON_CANCEL,      DIALOG_BUTTON_INVALID},
	{DIALOG_BUTTON_YES,    DIALOG_BUTTON_NO,          DIALOG_BUTTON_INVALID},
	{DIALOG_BUTTON_ACCEPT, DIALOG_BUTTON_ACCEPT_ONCE, DIALOG_BUTTON_DECLINE}
};

CDialogModal::CDialogModal(const int aWidth, const V2 aDialogPosition, const CString aText, EDialogButtonConfig aButtonConfig):
	m_size(0,0),
	m_position(aDialogPosition),
	m_text(aText),
	m_exitValue(DIALOG_BUTTON_INVALID),
	m_buttonConfig(aButtonConfig)
{
	// NULLify the button array
	for (int i = 0; i < MAX_MODAL_DLG_BUTTONS; i++)
	{
		m_pButtons[i] = NULL;
	}
	
	// Determine horizontal size
	float minWidth = getMinWidth(m_buttonConfig);
	m_size.x = aWidth < minWidth ? minWidth : aWidth;

	V2 text_size(0, 0);
	V2 button_size(BUTTON_WIDTH, BUTTON_HEIGHT);

	// If text is provided, determine text_size
	if (!m_text.IsEmpty()) {
		m_text = CFont::GetActiveUiFont()->GetWrappedString(m_text, m_size.x - WINDOW_BORDER*2 - WINDOW_PADDING*2);
		text_size = CFont::GetActiveUiFont()->GetStringSize(m_text);
	}

	// Build vertical total size
	m_size.y = WINDOW_BORDER * 2 + WINDOW_PADDING * 2 + button_size.y;
	if (!m_text.IsEmpty()) {
		m_size.y += text_size.y + ITEM_PADDING;
	}	

	// Center the dialog box if position coord comes as negative
	if (m_position.x < 0) {
		m_position.x = CGfx::s_ScreenWidth / 2 - m_size.x / 2;
	}
	if (m_position.y < 0) {
		m_position.y = CGfx::s_ScreenHeight /2 - m_size.y / 2;
	}

	// Setup item positioning
	V2 item_position(m_position.x + WINDOW_BORDER+WINDOW_PADDING, m_position.y + WINDOW_BORDER+WINDOW_PADDING);
	if (!m_text.IsEmpty()) {
		item_position.y += text_size.y + ITEM_PADDING;
	}

	// Button positioning
	int buttonCount = getButtonCount(m_buttonConfig);
	float totalButtonWidth = BUTTON_WIDTH*buttonCount + ITEM_PADDING*(buttonCount - 1);
	float middleX = m_position.x + (m_size.x / 2);
	float offsetX = middleX - (totalButtonWidth / 2);
	m_pControlManager = new CControlManager();
	
	for (int i = 0; i < buttonCount; i++)
	{
		V2 button_pos(offsetX, item_position.y);
		m_pButtons[i] = new CControlButton(button_size, button_pos, s_buttonCaptions[m_buttonConfig][i]);
		offsetX += BUTTON_WIDTH + ITEM_PADDING;
		m_pControlManager->addControl(m_pButtons[i]);
	}

	m_pControlManager->selectControl(m_pButtons[0]);

	// Setup navigation between buttons
	linkButtons();
    
    s_modalList.push_back(this);
}

CDialogModal::~CDialogModal()
{
	SAFE_DELETE(m_pControlManager);
	
	for (TModalList::iterator it = s_modalList.begin(); it != s_modalList.end(); ++it)
	{
		if (*it == this)
		{
			s_modalList.erase(it);
			break;
		}
	}
}

void CDialogModal::Open()
{
	CRenderable::Register(CRenderable::RO_DLG_MODAL, CDialogModal::Render);
}

void CDialogModal::Close()
{
	while (s_modalList.empty() == false)
	{
		CDialogModal* dialog(s_modalList.back());
		delete dialog;
	}
	
	CRenderable::UnRegister(CRenderable::RO_DLG_MODAL, CDialogModal::Render);
}

void CDialogModal::Process()
{
	//
	//	Process all the dialogs
	//
	for (TModalList::iterator it = s_modalList.begin(); it != s_modalList.end(); ++it)
	{
		(*it)->ProcessDialog();
	}

	//
	//	Process the input for the top most dialog
	//
	if (s_modalList.size() > 0)
		s_modalList.back()->ProcessDialogInput();
}

void CDialogModal::Render()
{
	//
	//	Render all the dialogs
	//
	for (TModalList::iterator it = s_modalList.begin(); it != s_modalList.end(); ++it)
	{
		(*it)->RenderDialog();
	}
}

bool CDialogModal::HasDialog()
{
	return !s_modalList.empty();
}

void CDialogModal::RenderDialog()
{
	// Draw background
	CGfx::DrawQuad( V2( 0, 0 ), V2( CGfx::s_ScreenWidth, CGfx::s_ScreenHeight), CSkinManager::GetColor( "dialog_box", "overlay_color", 0x8F000000));

	// Draw the borders
    // Sort of cheating here, couldn't send bgcolor to none opaque, if we want that functionality
	// need to draw 4 seperate quads
	CGfx::DrawQuad( V2( m_position.x, m_position.y ),
					V2( m_size.x, m_size.y),
					CSkinManager::GetColor( "dialog_box", "border_color", 0xFF00FF00));

	// Draw menu background (overlay over the menu border quad)
	CGfx::DrawQuad( V2( WINDOW_BORDER + m_position.x, WINDOW_BORDER + m_position.y),
					V2( m_size.x - WINDOW_BORDER * 2, m_size.y - WINDOW_BORDER * 2),
					CSkinManager::GetColor( "dialog_box", "background_color", 0xFF000000));

	// Draw the text
	V2 text_position(m_position.x + WINDOW_BORDER + WINDOW_PADDING, m_position.y + WINDOW_BORDER + WINDOW_PADDING);
	CFont::GetActiveUiFont()->Print(m_text, text_position, CSkinManager::GetColor( "dialog_box", "text_color", 0xFFFFFFFF));
	
	// Render the inputs
	m_pControlManager->RenderControl();
}

void CDialogModal::ProcessDialog()
{
	for (int i = 0; i < getButtonCount(m_buttonConfig); i++)
	{
		if (m_pButtons[i]->getClicked())
		{
			m_exitValue = s_buttonValues[m_buttonConfig][i];
			break;
		}
	}

	m_pControlManager->ProcessControl();
}

void CDialogModal::ProcessDialogInput()
{
	m_pControlManager->ProcessControlInput();
}

int CDialogModal::getExitValue()
{
	return m_exitValue;
}

float CDialogModal::getMinWidth(EDialogButtonConfig aButtonConfig)
{
	switch (aButtonConfig)
	{
		case DIALOG_CONFIG_OK_ONLY:
			return MIN_WIDTH_ONE_BUTTON;
		case DIALOG_CONFIG_OK_CANCEL:
			return MIN_WIDTH_TWO_BUTTONS;
		case DIALOG_CONFIG_YES_NO:
			return MIN_WIDTH_TWO_BUTTONS;
		case DIALOG_CONFIG_ACCEPT_DECLINE:
			return MIN_WIDTH_THREE_BUTTONS;
		default:
			return 0;
	}
}

int CDialogModal::getButtonCount(EDialogButtonConfig aButtonConfig)
{
	switch (aButtonConfig)
	{
		case DIALOG_CONFIG_OK_ONLY:
			return 1;
		case DIALOG_CONFIG_OK_CANCEL:
			return 2;
		case DIALOG_CONFIG_YES_NO:
			return 2;
		case DIALOG_CONFIG_ACCEPT_DECLINE:
			return 3;
		default:
			return 0;
	}
}

void CDialogModal::linkButtons()
{
	if (MAX_MODAL_DLG_BUTTONS < 2)
		return;
	
	int count = 0;
	CControlButton *prev, *next;
	
	while ((count + 1) < MAX_MODAL_DLG_BUTTONS)
	{
		prev = m_pButtons[count];
		next = m_pButtons[count+1];
		
		if (next == NULL)
			return;
			
		prev->m_pRight = next;
		next->m_pLeft = prev;
		count++;
	}
}
