/*******************************************************************************

  PSPoste - Wireless Email Client for the Sony PSP
  Copyright (C) T Swann, Benjamin Biron, Sajeeth Cherian, David Cummings

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

  Module :	CFont.cpp

  Description :	

  Last Modified $Date: 2008/08/31 03:50:11 $

  $Revision: 1.4 $

*******************************************************************************/

//******************************************************************************
//   Include Files
//******************************************************************************
#include "CFont.h"
#include "CFrameWork.h"
#include "CTextureManager.h"
#include "CGfx.h"
#include "CSkinManager.h"
#include "CCachedFont.h"
#include <CTextAsciiRenderTable.h>

//******************************************************************************
//   Local Macros
//******************************************************************************

//******************************************************************************
//   Local Constants
//******************************************************************************
static const float  CURSOR_WIDTH( 1.f ); // Width of the text cursor

//******************************************************************************
//   Static Prototypes
//******************************************************************************

//******************************************************************************
//   Global Variables
//******************************************************************************

//******************************************************************************
//   Static Variables
//******************************************************************************
CFont* CFont::s_pActiveUiFont(NULL);
CFont* CFont::s_pActiveTextFont(NULL);

//******************************************************************************
//   Class Definition
//******************************************************************************

//******************************************************************************
//	
//******************************************************************************
bool CFont::Open()
{
	int uiFontSize = (int)CSkinManager::GetFloat("font", "ui_size", 12.f);
	int textFontSize = (int)CSkinManager::GetFloat("font", "text_size", 12.f);
	CString uiFontPath = CSkinManager::GetUiFontPath();
	CString textFontPath = CSkinManager::GetTextFontPath();
	
	s_pActiveUiFont = CTextureManager::CreateCachedFont(uiFontPath, uiFontSize);
	if (s_pActiveUiFont == NULL)
	{
		CLog::Error("Failed to create the UI font\n");
		return false;
	}

	s_pActiveTextFont = CTextureManager::CreateCachedFont(textFontPath, textFontSize);
	if (s_pActiveTextFont == NULL)
	{
		CLog::Error("Failed to create the Text font\n");
		return false;
	}
	
	return true;
}

//******************************************************************************
//	
//******************************************************************************
void CFont::Close()
{
	SAFE_DELETE(s_pActiveUiFont);
	SAFE_DELETE(s_pActiveTextFont);
}

//******************************************************************************
//	
//******************************************************************************
CFont* CFont::GetActiveUiFont()
{
	ASSERT(s_pActiveUiFont != NULL, "Trying to access a NULL font");
	return s_pActiveUiFont;
}

//******************************************************************************
//	
//******************************************************************************
CFont* CFont::GetActiveTextFont()
{
	ASSERT(s_pActiveTextFont != NULL, "Trying to access a NULL font");
	return s_pActiveTextFont;
}

//******************************************************************************
//	
//******************************************************************************
CFont::CFont(u32 fontSize) : CTexture(CTexture::rgba32),
	m_fontSize(fontSize),
	m_ctextRenderTable(NULL),
	m_fontMetrics(NULL),
	m_metricTableSize(0),
	m_lineHeight(0)
{

}

//******************************************************************************
//	
//******************************************************************************
CFont::~CFont()
{
	SAFE_DELETE(m_ctextRenderTable);
	
	// m_fontMetrics is allocated in the TTFReader with memalign, so we must use
	// free on it
	if (m_fontMetrics)
		free(m_fontMetrics);
}

//******************************************************************************
//	
//******************************************************************************
void CFont::InitCTextRenderTable()
{
	// The CText "advance" is different than the FreeType advance. The FreeType
	// advance is from beginning of glyph to beginning of next glyph. The CText
	// "advance" is from end of glyph to beginning of next glyph.
	u32 ctextAdvance;
	
	SAFE_DELETE(m_ctextRenderTable);
	
	m_ctextRenderTable = new CTextAsciiRenderTable();
	
	// We only support ascii for now
	for (u32 i = 0; i < 256; i++)
	{
		ctextAdvance = m_fontMetrics[i].xadvance - m_fontMetrics[i].width;
		m_ctextRenderTable->set(i, m_fontMetrics[i].width, ctextAdvance);
	}
	
	// Let's not forget the space character
	//s_ctext_rtable->set(' ', (size_t)m_Size.x, 0);
}

//******************************************************************************
//	
//******************************************************************************
void CFont::Print( const CString & string, const V2 & pos, ARGB color)
{
	const char *			p_char( string );
	const V2				orig_pos( static_cast< float >( static_cast< s32 >( pos.x ) ), static_cast< float >( static_cast< s32 >( pos.y ) ) );
	V2						text_pos( orig_pos );
	u32						sprite_count( 0 );
	sVertexTexturedColor *	p_vert;
	sVertexTexturedColor *	p_dst;

	CGfx::GetPolyList( 2 * string.Length(), &p_vert );

	p_dst = p_vert;

	for ( s32 i = 0; i < string.Length(); ++i )
	{
		switch ( *p_char )
		{
		case ' ':
			{
				text_pos.x += m_fontMetrics[' '].xadvance;
			}
			break;

		case '\n':
			{
				text_pos.x = orig_pos.x;
				text_pos.y += GetHeight() + 1;
			}
			break;

		default:
			{
				const Metrics p_glyph_data(m_fontMetrics[(u32)*p_char]);

				p_dst->uv.x = p_glyph_data.x;
				p_dst->uv.y = p_glyph_data.y;
				p_dst->color = color;
				p_dst->pos.x = text_pos.x;
				p_dst->pos.y = text_pos.y;
				p_dst->pos.z = 0.f;
				++p_dst;

				p_dst->uv.x = p_glyph_data.x + p_glyph_data.width;
				p_dst->uv.y = p_glyph_data.y + p_glyph_data.height;
				p_dst->color = color;
				p_dst->pos.x = text_pos.x + p_glyph_data.width;
				p_dst->pos.y = text_pos.y + p_glyph_data.height;
				p_dst->pos.z = 0.f;
				++p_dst;

				text_pos.x += p_glyph_data.xadvance;

				++sprite_count;
			}
			break;
		}

		p_char++;
	}

	CGfx::DrawSprite(this, p_vert, 2 * sprite_count);
}

bool CFont::Print(const CText* text, const int aLineNum, const V2 & size, const V2 & pos, ARGB color, const bool aShowCursor)
{
	// Returns true if at the end of the CText
	const V2				orig_pos( static_cast< float >( static_cast< s32 >( pos.x ) ), static_cast< float >( static_cast< s32 >( pos.y ) ) );
	V2						text_pos( orig_pos );
	u32						sprite_count( 0 );
	sVertexTexturedColor *	p_vert;
	sVertexTexturedColor *	p_dst;
	sVertexTexturedColor *	p_wrapDst;
	std::vector<sVertexTexturedColor *> vertList;
	float					x_out(0.f);
	float					y_out(0.f);
	bool					done = false; // Will be true iff we reach the end of the CText

	int						curLineNum = 0; // The line number we are currently parsing
	bool					outputChar = false; // Should the character be added to the vertex list?
	u32						curListIdx = 0; // The index of the array within the list that is currently being filled
	u32						markedListIdx = 0; // The mark index within the list of vertex arrays
	u32						markedVertIdx = 0; // The mark index within the vertex array
	bool					markValid = false; // Is the mark corresponding to the last word valid?
	bool					backBuffer = false; // Should we be filling the "back buffer"?
	bool					foundCursor = false; // Has the cursor been found before the end?

	// Because we can't know exactly how much text will be rendered, we allocate
	// the GU memory in chunks of approximately 1 line of text.
	u32 guAllocCnt = static_cast<u32>(size.x / m_fontMetrics[' '].xadvance);
	
	CGfx::GetPolyList( 2 * guAllocCnt, &p_vert );
	p_dst = p_vert;
	
	CTextIterator it(text->begin());

	//CString out;
	//out.Printf("aLineNum = %d", aLineNum);
	//ASSERT(0, out.GetPtr());

	// The done variable is toggled when rendering is done up to size.y
	while ( it.hasNext() && !done )
	{
		CTextChar& ch = ++it;
		outputChar = (curLineNum >= aLineNum);
		
		if (aShowCursor && ch.m_has_cursor)
		{
			foundCursor = true;
			CGfx::DrawQuad( V2( text_pos.x, text_pos.y - 1.f),
					        V2( CURSOR_WIDTH, GetHeight()),
					        CSkinManager::GetColor( "text_box", "cursor_color", 0xFFFFFFFF));
		}
		
		switch ( ch.m_char )
		{
		case ' ':
			{
				text_pos.x += m_fontMetrics[' '].xadvance;
				x_out += m_fontMetrics[' '].xadvance;
				
				// Note that wheter or not outputChar is true, we need to mark
				// the beginning of the word, this is because the word may end
				// up being wrapped on the first displayed line.
				markValid = true;
				markedListIdx = curListIdx;
				
				if (!outputChar)
				{
					// We enable the "back buffer" to catch the wrapped word
					backBuffer = true;
					sprite_count = 0;
					p_dst = p_vert;
				}
				
				markedVertIdx = sprite_count;
			}
			break;

		case '\r':
			// Just discard the carriage returns
			break;

		case '\n':
			{
				text_pos.x = orig_pos.x;
				x_out = 0;
				
				if (outputChar)
				{
					text_pos.y += GetHeight() + 1;
					y_out += GetHeight() + 1;
					markValid = false;
				}
				else
				{
					// The back buffer must be "flushed"
					backBuffer = false;
					sprite_count = 0;
					p_dst = p_vert;
				}
				
				curLineNum++;
				//CString out;
				//out.Printf("A: %d", curLineNum);
				//ASSERT1(0, out.GetPtr());
			}
			break;

		default:
			{
				const Metrics p_glyph_data(m_fontMetrics[ch.m_char]);
				
				//CString out;
				//out.Printf("%d '%c'", outputChar, ch.m_char);
				//ASSERT1(0, out.GetPtr());
				
				// Have we used up the allocated space for the sprite list?
				if (sprite_count == guAllocCnt && outputChar)
				{
					// We've used up all the allocated list, let's enqueue it
					// then get some more memory.
					vertList.push_back(p_vert);
					
					sprite_count = 0;
					CGfx::GetPolyList( 2 * guAllocCnt, &p_vert );
					p_dst = p_vert;
					
					curListIdx++;
				}
				
				// Let's see if this character fits on the line, if not... we
				// must wrap.
				if ((x_out + p_glyph_data.width) > size.x)
				{
					//ASSERT(0, "Hello?");
					// It doesn't fit; let's wrap, starting at the last word
					// boundary.
					text_pos.x = orig_pos.x;
					x_out = 0;
					
					if (outputChar)
					{
						text_pos.y += GetHeight() + 1;
						y_out += GetHeight() + 1;
					}
					
					// The basic idea is to go back to the vertex corresponding
					// to the first character of the last word situated after a
					// space, and start modifying the vertices to start on the
					// new line.
					if (markValid && (outputChar || backBuffer))
					{
						//ASSERT1(0, "markValid == true");

						while (1)
						{
							//ASSERT1(0, "Begin loop");
							// We're at the end of this vertex "array", so we
							// should go to the next array in the vert list
							if (markedVertIdx == guAllocCnt)
							{
								//ASSERT1(0, "At end of list");
								markedVertIdx = 0;
								
								if (++markedListIdx > curListIdx)
									break;
							}
							
							// This would mean the vertices haven't been added to the list yet
							if (markedListIdx == curListIdx)
							{
								//ASSERT1(0, "In current list");
								p_wrapDst = p_vert + (markedVertIdx*2);
								
								if (p_wrapDst == p_dst)
									break;
								
								//CString out;
								//out.Printf("%x %x %x", p_vert, p_wrapDst, p_dst);
								//ASSERT(0, out.GetPtr());
							}
							else
							{
								//ASSERT1(0, "In past list");
								//CString out;
								//out.Printf("curListIdx = %d, markedListIdx = %d, markedVertIdx = %d", curListIdx, markedListIdx, markedVertIdx);
								//ASSERT1(0, out.GetPtr());
								p_wrapDst = vertList[markedListIdx] + (markedVertIdx*2);
								//out.Printf("p_wrapDst = %x, p_dst = %x", p_wrapDst, p_dst);
								//ASSERT1(0, out.GetPtr());
							}
								
							// Update the positions of the existing vertices
							float oldInc = (p_wrapDst+1)->pos.x - p_wrapDst->pos.x;
							p_wrapDst->pos.x = text_pos.x;
							p_wrapDst->pos.y = text_pos.y;
							p_wrapDst++;
							
							p_wrapDst->pos.x = text_pos.x + oldInc;
							p_wrapDst->pos.y = text_pos.y + GetHeight();
							
							text_pos.x += oldInc + 1;
							x_out += oldInc + 1;
							
							markedVertIdx++;
						}
					}
					
					markValid = false;
					curLineNum++;
					
					//ASSERT1(0, "Loop done");
					//CString out;
					//out.Printf("B: %d", curLineNum);
					//ASSERT1(0, out.GetPtr());
				}

				// Let's make sure we haven't wrapped past the area size. Note
				// that if we're at the first line, we should render no matter
				// what!
				if ((y_out + (GetHeight() + 1)) > size.y && curLineNum > 0)
				{
					done = true;
					break;
				}

				if (outputChar || backBuffer)
				{
					// Let's add the vertices to the list
					p_dst->uv.x = p_glyph_data.x;
					p_dst->uv.y = p_glyph_data.y;
					p_dst->color = color;
					p_dst->pos.x = text_pos.x;
					p_dst->pos.y = text_pos.y;
					p_dst->pos.z = 0.f;
					++p_dst;
	
					p_dst->uv.x = p_glyph_data.x + p_glyph_data.width;
					p_dst->uv.y = p_glyph_data.y + p_glyph_data.height;
					p_dst->color = color;
					p_dst->pos.x = text_pos.x + p_glyph_data.width;
					p_dst->pos.y = text_pos.y + p_glyph_data.height;
					p_dst->pos.z = 0.f;
					++p_dst;
	
					text_pos.x += p_glyph_data.xadvance;
	
					++sprite_count;
				}
				
				// Note that x_out has to be updated whether or not outputChar
				// is true, this is to cover the case of the first line output
				// being due to wrapped text as opposed to a newline.
				x_out += p_glyph_data.xadvance;

			}
			
			break;
		}
	}
	
	if (aShowCursor && !foundCursor)
	{
		if (!done || (++it).m_has_cursor)
		{
			CGfx::DrawQuad( V2( text_pos.x, text_pos.y - 1.f),
					        V2( CURSOR_WIDTH, GetHeight()),
					        CSkinManager::GetColor( "text_box", "cursor_color", 0xFFFFFFFF));
		}
	}

	// Let's send the vertices to the GU. Note that each list in the vertList
	// will be full.
	std::vector<sVertexTexturedColor*>::iterator vertIt = vertList.begin();
	while (vertIt != vertList.end())
	{
		/*
		for (unsigned int i = 0; i < (guAllocCnt*2); i++)
		{
			sVertexTexturedColor* aVert = (*vertIt)+i;
			CLog::Message("-- 1 -- [%x]\n", aVert);
			CLog::Message("u = %f\n", aVert->u);
			CLog::Message("v = %f\n", aVert->v);
			CLog::Message("c = %x\n", aVert->color.color);
			CLog::Message("x = %f\n", aVert->x);
			CLog::Message("y = %f\n", aVert->y);
			CLog::Message("z = %f\n", aVert->z);
			CLog::Message("\n");
		}
		*/
		
		CGfx::DrawSprite( this, *vertIt, 2 * guAllocCnt );
		vertIt++;
	}

	// The last list may not have been added to the vertList because it may not
	// have reached the full point. We must send these vertices off also!
	if (sprite_count != 0)
	{
		/*
		for (unsigned int i = 0; i < (sprite_count*2); i++)
		{
			sVertexTexturedColor* aVert = p_vert+i;
			CLog::Message("-- 2 -- [%x]\n", aVert);
			CLog::Message("u = %f\n", aVert->u);
			CLog::Message("v = %f\n", aVert->v);
			CLog::Message("c = %x\n", aVert->color.color);
			CLog::Message("x = %f\n", aVert->x);
			CLog::Message("y = %f\n", aVert->y);
			CLog::Message("z = %f\n", aVert->z);
			CLog::Message("\n");
		}
		*/
		
		CGfx::DrawSprite( this, p_vert, 2 * sprite_count );
	}
		
	// Note that if done is true, it means we broke from the loop before
	// reaching the end of the CText.
	return (!done);
}

//******************************************************************************
//	
//******************************************************************************
void	CFont::Print(CTextPtr aStart, size_t aCount, const V2 & pos, ARGB color)
{
	V2						orig_pos( static_cast< float >( static_cast< s32 >( pos.x ) ), static_cast< float >( static_cast< s32 >( pos.y ) ) );
	V2						text_pos(orig_pos);
	u32						sprite_count( 0 );
	sVertexTexturedColor *	p_vert;
	sVertexTexturedColor *	p_dst;

	CGfx::GetPolyList( 2 * aCount, &p_vert );

	p_dst = p_vert;

	for ( u32 i = 0; i < aCount && aStart.valid(); ++i )
	{
		switch ( *aStart )
		{
		case ' ':
			{
				text_pos.x += m_fontMetrics[' '].xadvance;
			}
			break;

		case '\r':
			// Just ignore carriage returns
			break;

		case '\n':
			{
				text_pos.x = orig_pos.x;
				text_pos.y += GetHeight() + 1;
			}
			break;

		default:
			{
				const Metrics p_glyph_data(m_fontMetrics[*aStart]);

				p_dst->uv.x = p_glyph_data.x;
				p_dst->uv.y = p_glyph_data.y;
				p_dst->color = color;
				p_dst->pos.x = text_pos.x;
				p_dst->pos.y = text_pos.y;
				p_dst->pos.z = 0.f;
				++p_dst;

				p_dst->uv.x = p_glyph_data.x + p_glyph_data.width;
				p_dst->uv.y = p_glyph_data.y + p_glyph_data.height;
				p_dst->color = color;
				p_dst->pos.x = text_pos.x + p_glyph_data.width;
				p_dst->pos.y = text_pos.y + p_glyph_data.height;
				p_dst->pos.z = 0.f;
				++p_dst;

				text_pos.x += p_glyph_data.xadvance;

				++sprite_count;
			}
			break;
		}

		++aStart;
	}

	CGfx::DrawSprite( this, p_vert, 2 * sprite_count );
}

//******************************************************************************
//	
//******************************************************************************
V2	CFont::GetStringSize( const CString & string ) const
{
	const char *	p_char( string );
	V2				size( 0.f, GetHeight() );
	float			last_x( 0.f );

	for ( s32 i = 0; i < string.Length(); ++i )
	{
		switch ( *p_char )
		{
		case ' ':
			{
				size.x += m_fontMetrics[' '].xadvance;
			}
			break;

		case '\n':
			{
				if ( last_x < size.x )
				{
					last_x = size.x;
				}

				size.x = 0.f;
				size.y += GetHeight() + 1;
			}
			break;

		default:
			{
				const Metrics p_glyph_data(m_fontMetrics[(u32)*p_char]);
				size.x += p_glyph_data.xadvance;
			}
			break;
		}

		p_char++;
	}

	if ( last_x > size.x )
	{
		size.x = last_x;
	}

	return size;
}

//******************************************************************************
//	
//******************************************************************************
CString	CFont::GetWrappedString( const CString & string, float width ) const
{
	const char *p_char( string );
	const char *p_begin = p_char;
	const char *p_lastSpace = NULL;

	CString widthCompare("");
	CString returnString("");
	CString temp("");

	for ( s32 i = 0; i < string.Length(); ++i )
	{
		widthCompare = CString("");
		switch ( *p_char )
		{
			case ' ':
				{
					p_lastSpace = p_char;
				}
				break;
			default:
				{
					// copy characters into widthCompare CString
					for(const char *p_loop = p_begin; p_loop != p_char; p_loop++) {
						widthCompare += temp.Printf("%c", *p_loop);
					}
					// string exceeds width
					if (GetStringSize(widthCompare).x > width) {
						//There is a previous space in the string
						if (p_lastSpace != NULL) {
							for(const char *p_loop = p_begin; p_loop != p_lastSpace; p_loop++) {
								returnString += temp.Printf("%c", *p_loop);
							}
							returnString += "\n";
							p_begin = p_lastSpace + 1;
						} else {
							widthCompare = CString("");
							for(const char *p_loop = p_begin; p_loop != p_char; p_loop++) {
								if (GetStringSize(widthCompare +  temp.Printf("%c", *p_loop)).x > width) {
									returnString += widthCompare;
									returnString += "\n";
									p_begin = p_loop;
									break;
								}
								widthCompare += temp.Printf("%c", *p_loop);
							}
							p_begin = ++p_char;
							continue;
						}
						p_lastSpace = NULL;
					} 
				}
				break;
		}
		p_char++;
	}

	// Copy over the last line
	for(const char *p_loop = p_begin; p_loop != p_char; p_loop++) {
		returnString += temp.Printf("%c", *p_loop);
	}
	return returnString;
}

//******************************************************************************
//	
//******************************************************************************
void CFont::SetMetrics(Metrics* metrics, u32 metricTableSize)
{
	// m_fontMetrics is allocated in the TTFReader with memalign, so we must use
	// free on it
	if (m_fontMetrics != NULL)
		free(m_fontMetrics);
		
	m_fontMetrics = metrics;
	m_metricTableSize = metricTableSize;
	
	InitCTextRenderTable();
}

//******************************************************************************
//	
//******************************************************************************
void CFont::SetHeight(u32 lineHeight)
{
	m_lineHeight = lineHeight;
}

//*******************************  END OF FILE  ********************************
