/*******************************************************************************

  PSPoste - Wireless Email Client for the Sony PSP
  Copyright (C) T Swann, Benjamin Biron, Sajeeth Cherian, David Cummings

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

  Module :	CInput.cpp

  Description :	

  Last Modified $Date: 2008/09/24 01:54:31 $

  $Revision: 1.5 $

*******************************************************************************/

//******************************************************************************
//   Include Files
//******************************************************************************
#include "CTypes.h"
#include <pspctrl.h>
#include "CInput.h"
#include "CFrameWork.h"
#include "CConfigFile.h"
#include "CRenderable.h"

#include "CKeyboardGUI.h"
#include "CKeyboardGUIAnalog1.h"
#include "CKeyboardIrda.h"
#include "CKeyboardDanzeff.h"
#include "CKeyboardPsprint.h"
#include "CKeyboardPspirkeyb.h"
#include "CKeyboardNav.h"
#include "CKeyboardNavGUIAnalog1.h"
#include "CKeyboardNavIrda.h"
#include "CKeyboardNavWrapper.h"
//******************************************************************************
//   Local Macros
//******************************************************************************

//******************************************************************************
//   Local Constants
//******************************************************************************

//******************************************************************************
//   Static Prototypes
//******************************************************************************

//******************************************************************************
//   Global Variables
//******************************************************************************

//******************************************************************************
//   Static Variables
//******************************************************************************
static SceCtrlData	s_PadData;
static bool			s_bPrevButton[ CInput::MAX_BUTTONS ];
static bool			s_bLastButton[ CInput::MAX_BUTTONS ];
static s32			s_nButtonRepeat[ CInput::MAX_BUTTONS ];
static const u32	s_ButtonMap[ CInput::MAX_BUTTONS ] =
{
	PSP_CTRL_SELECT,	// SELECT
	PSP_CTRL_START,		// START
	PSP_CTRL_UP,		// UP
	PSP_CTRL_RIGHT,		// RIGHT
	PSP_CTRL_DOWN,		// DOWN
	PSP_CTRL_LEFT,		// LEFT
	PSP_CTRL_LTRIGGER,	// LTRIGGER
	PSP_CTRL_RTRIGGER,	// RTRIGGER
	PSP_CTRL_TRIANGLE,	// TRIANGLE
	PSP_CTRL_CIRCLE,	// CIRCLE
	PSP_CTRL_CROSS,		// CROSS
	PSP_CTRL_SQUARE,	// SQUARE
	PSP_CTRL_HOME,		// HOME
	PSP_CTRL_HOLD,		// HOLD
	PSP_CTRL_NOTE,		// NOTE
};

// Repeat values
const s32	CInput::PAD_SLOW_REPEAT = 8;
const s32	CInput::PAD_FAST_REPEAT = 2;
const s32	CInput::MAX_REPEAT_COUNT = 1;

// CKeyboard stuff
CKeyboard* CInput::s_keyboard = NULL;
CKeyboardNav* CInput::s_keyboardNav = NULL;

//******************************************************************************
//   Class Definition
//******************************************************************************

//******************************************************************************
//	
//******************************************************************************

//******************************************************************************
//	
//******************************************************************************
bool	CInput::Open()
{
	for ( u32 b = 0; b < MAX_BUTTONS; ++b )
	{
		s_bPrevButton[ b ] = false;
		s_bLastButton[ b ] = false;
	}

	sceCtrlSetSamplingCycle( 0 );
	sceCtrlSetSamplingMode( PSP_CTRL_MODE_ANALOG );
	
	SetKeyboardFromName(CConfigFile::s_pSettings->keyboard, CConfigFile::s_pSettings->analogNav);
	
	CRenderable::Register(CRenderable::RO_KEYBOARD, RenderKeyboard);
	
	return true;
}

//******************************************************************************
//	
//******************************************************************************
void	CInput::Close()
{
	CKeyboardNavWrapper::SetActiveNav(NULL);
	CRenderable::UnRegister(CRenderable::RO_KEYBOARD, RenderKeyboard);
	SAFE_DELETE(s_keyboard);
	SAFE_DELETE(s_keyboardNav);
}

//******************************************************************************
//	
//******************************************************************************
void	CInput::Process()
{
	for ( u32 i = 0; i < MAX_BUTTONS; ++i )
	{
		s_bPrevButton[ i ] = ( s_PadData.Buttons & s_ButtonMap[ i ] );
	}

	for ( u32 i = 0; i < MAX_BUTTONS; ++i )
	{
		if ( IsButtonDown( static_cast< eButton >( i ) ) == true )
		{
			s_nButtonRepeat[ i ]++;
		}
		else
		{
			s_nButtonRepeat[ i ] = 0;
		}
	}

	sceCtrlReadBufferPositive( &s_PadData, 1 );
	
	if (s_keyboard)
		s_keyboard->process(&s_PadData);
		
	if (s_keyboardNav)
		s_keyboardNav->process(&s_PadData);
}

//******************************************************************************
//	
//******************************************************************************
bool	CInput::IsButtonDown( eButton button )
{
	return ( s_PadData.Buttons & s_ButtonMap[ button ] );
}

//******************************************************************************
//	
//******************************************************************************
bool	CInput::IsButtonClicked( eButton button )
{
	if ( IsButtonDown( button ) == true )
	{
		if ( s_bPrevButton[ button ] == false )
		{
			s_bPrevButton[ button ] = true;

			return true;
		}
	}

	return false;
}

//******************************************************************************
//	
//******************************************************************************
bool	CInput::IsButtonReleased( eButton button )
{
 	if ( IsButtonDown( button ) == true )
	{
		if ( s_bPrevButton[ button ] == false )
		{
			s_bPrevButton[ button ] = true;
			s_bLastButton[button] = true;
		}
	}	
	if (s_bLastButton[button] && IsButtonDown( button ) == false )
	{
	   	s_bLastButton[button] = false;
		return true;
	}
	return false;
}

//******************************************************************************
//	
//******************************************************************************
bool	CInput::IsButtonRepeat( eButton button, s32 repeat )
{
	if ( IsButtonDown( button ) == true )
	{
		return ( ( s_nButtonRepeat[ button ] % repeat ) == 0 );
	}

	return false;
}

//******************************************************************************
//	
//******************************************************************************
V2	CInput::GetAnalogStick()
{
	float	analog_x( static_cast< float >( s_PadData.Lx ) );
	float	analog_y( static_cast< float >( s_PadData.Ly ) );

	analog_x *= 2.f;
	analog_y *= 2.f;

	analog_x /= 255.f;
	analog_y /= 255.f;

	analog_x -= 1.f;
	analog_y -= 1.f;

	return V2( analog_x, analog_y );
}

//******************************************************************************
//	
//******************************************************************************
bool	CInput::IsKeyboardPressed()
{
 	if (s_keyboard)
 		return (s_keyboard->isKeyboardPressed() != CKeyboard::eKeyboardKeyPressed_None);

	return false;
}

//******************************************************************************
//	
//******************************************************************************
char CInput::GetKeyboardCharKey()
{
 	if (s_keyboard)
 		return (char)(s_keyboard->getCharKey() & 0xff);
 	
 	return 0;
}

//******************************************************************************
//	
//******************************************************************************
void CInput::SetKeyboardFromName(CString aKeyboardName, bool analogNav)
{
	CKeyboardNavWrapper::SetActiveNav(NULL);
	SAFE_DELETE(s_keyboard);
	SAFE_DELETE(s_keyboardNav);
	
	if (aKeyboardName == "IR Keyboard (Stowaway Legacy)")
	{
		s_keyboard = new CKeyboardIrda();
		s_keyboardNav = new CKeyboardNavIrda(s_keyboard);
	}
	else if (aKeyboardName == "IR Keyboard (pspirkeyb)")
	{
		s_keyboard = new CKeyboardPspirkeyb();
		s_keyboardNav = new CKeyboardNavIrda(s_keyboard);		
	}
	else
	{
		if (aKeyboardName == "GUI Keyboard")
			s_keyboard = new CKeyboardGUI();
		else if (aKeyboardName == "GUI Keyboard Analog1")
			s_keyboard = new CKeyboardGUIAnalog1();
		else if (aKeyboardName == "p-sprint")
			s_keyboard = new CKeyboardPsprint();
		else if (aKeyboardName == "Danzeff")
			s_keyboard = new CKeyboardDanzeff();
		
		if (analogNav)
			s_keyboardNav = new CKeyboardNavGUIAnalog1(s_keyboard);
		else
			s_keyboardNav = new CKeyboardNav(s_keyboard);
	}
	
	CKeyboardNavWrapper::SetActiveNav(s_keyboardNav);
}

//******************************************************************************
//	
//******************************************************************************
void CInput::RenderKeyboard()
{
	if (s_keyboard)
	{
		s_keyboard->render();
	}
}

//******************************************************************************
//	
//******************************************************************************
void CInput::EnableKeyboard()
{
	if (s_keyboard)
	{
		s_keyboard->enableRender();
		s_keyboard->enableInput();
	}
}

//******************************************************************************
//	
//******************************************************************************
void CInput::DisableKeyboard()
{
	if (s_keyboard)
	{
		s_keyboard->disableRender();
		s_keyboard->disableInput();
	}
}

//******************************************************************************
//	
//******************************************************************************
u16 CInput::GetKeyboardControlKey()
{
	if (s_keyboard)
		return s_keyboard->getControlKey();
	
	return 0;
}

//******************************************************************************
//	
//******************************************************************************
void CInput::ResetKeyboard()
{
	if (s_keyboard)
		s_keyboard->reset();
}

//******************************************************************************
//	
//******************************************************************************
bool CInput::IsKeyboardEnabled()
{
	if (s_keyboard)
		return s_keyboard->isInputEnabled();
	
	return false;
}

//******************************************************************************
//	
//******************************************************************************
bool CInput::ShouldDisableKeyboard()
{
	if (s_keyboard)
		return s_keyboard->shouldDisable();
	
	return true;
}

//*******************************  END OF FILE  ********************************
