/*******************************************************************************

  PSPoste - Wireless Email Client for the Sony PSP
  Copyright (C) Benjamin Biron, Sajeeth Cherian, David Cummings

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

  Module :	CLog.cpp

  Description :	

  Last Modified $Date: 2008/10/01 03:02:57 $

  $Revision: 1.4 $

*******************************************************************************/

//******************************************************************************
//   Include Files
//******************************************************************************
#include "CLog.h"
#include <time.h>

//******************************************************************************
//   Defines
//******************************************************************************
//#define CLOG_SAFE_MODE
#define CLOG_ENABLE_LOCKS
#define CLOG_ENABLE_TIMESTAMPS

#ifdef CLOG_ENABLE_LOCKS
#define CHECK_LOCK_STATE if (m_sema <= 0) return;
#define CHECK_LOCK_STATE_RETURN_ZERO if (m_sema <= 0) return 0;
#define CHECK_LOCK_STATE_RETURN_NEG if (m_sema <= 0) return -1;
#define LOCK lock();
#define UNLOCK unlock();
#else
#define CHECK_LOCK_STATE 
#define CHECK_LOCK_STATE_RETURN_ZERO 
#define CHECK_LOCK_STATE_RETURN_NEG 
#define LOCK 
#define UNLOCK 
#endif

//******************************************************************************
//   Static Variables
//******************************************************************************
FILE* CLog::s_out(NULL);
FILE* CLog::s_err(NULL);
SceUID CLog::m_sema(0);

//******************************************************************************
//   Class implementations
//******************************************************************************
int CLog::Open(int aOverwrite)
{
	int toRet;
#ifdef CLOG_ENABLE_LOCKS
	m_sema = sceKernelCreateSema("CLog_Sema", 0, 1, 1, NULL);
#endif
	
	CHECK_LOCK_STATE_RETURN_ZERO
	
	LOCK
	
	toRet = SysOpen(aOverwrite);
	
	UNLOCK
	return toRet;
}

void CLog::Close()
{
	CHECK_LOCK_STATE
	
	LOCK
	
#ifdef CLOG_SAFE_MODE
	// Don't need to do anything here
#else
	SysClose();
#endif

	UNLOCK
	
#ifdef CLOG_ENABLE_LOCKS
	sceKernelDeleteSema(m_sema);
	m_sema = 0;
#endif
}

void CLog::Flush()
{
	CHECK_LOCK_STATE
	
#ifdef CLOG_SAFE_MODE
	// Flushing in safe mode is moot
#else
	LOCK

	SysClose();
	SysOpen(0);
	
	UNLOCK
#endif
}

int CLog::IsOpen()
{
	int toRet;
	
	LOCK
	
	toRet = SysIsOpen();
	
	UNLOCK
	return toRet;
}

int CLog::Message(const char* aMsg, ...)
{
	CHECK_LOCK_STATE_RETURN_NEG
		
	LOCK
	
#ifdef CLOG_SAFE_MODE
	s_out = fopen("out.log", "a");
#endif

	if (s_out > 0)
	{
		int toRet;
		va_list ap;
		
#ifdef CLOG_ENABLE_TIMESTAMPS
		SysTimestamp(s_out);
#endif
		va_start(ap, aMsg);
		toRet = vfprintf(s_out, aMsg, ap);
		va_end(ap);
		fflush(s_out);
#ifdef CLOG_SAFE_MODE
		fclose(s_out);
		s_out = NULL;
#endif
		UNLOCK
		return toRet;
	}
	
	UNLOCK
	return -1;
}

int CLog::MessageRaw(const char* aMsg, int len)
{
	CHECK_LOCK_STATE_RETURN_NEG
		
	LOCK
	
	int toRet;
	
#ifdef CLOG_SAFE_MODE
	s_out = fopen("out.log", "a");
	if (s_out > 0)
	{
 #ifdef CLOG_ENABLE_TIMESTAMPS
		SysTimestamp(s_out);
 #endif
		toRet = fwrite(aMsg, sizeof(char), len, s_out);
		fclose(s_out);
		s_out = NULL;
		UNLOCK
		return toRet;
	}
#else
	if (s_out > 0)
	{
 #ifdef CLOG_ENABLE_TIMESTAMPS
		SysTimestamp(s_out);
 #endif
		toRet = fwrite(aMsg, sizeof(char), len, s_out);
		UNLOCK
		return toRet;
	}
#endif
	
	UNLOCK
	return -1;
}

int CLog::Debug(const char* aMsg, ...)
{
#ifdef _DEBUG

	CHECK_LOCK_STATE_RETURN_NEG

	LOCK

 #ifdef CLOG_SAFE_MODE
    s_out = fopen("out.log", "a");
 #endif

	if (s_out > 0)
	{
		int toRet;
		va_list ap;
		
 #ifdef CLOG_ENABLE_TIMESTAMPS
		SysTimestamp(s_out);
 #endif
		
		va_start(ap, aMsg);
		fwrite("[Debug]: ", sizeof(char), 9, s_out);
		toRet = vfprintf(s_out, aMsg, ap);
		fwrite("\n", sizeof(char), 1, s_out);
		va_end(ap);
		fflush(s_out);
 #ifdef CLOG_SAFE_MODE
        fclose(s_out);
        s_out = NULL;
 #endif
		UNLOCK
		return toRet;
	}
	
	UNLOCK
	return -1;
#else	// #ifdef _DEBUG
	return 0;
#endif	// #ifdef _DEBUG
}

int CLog::DebugRaw(const char* aMsg, int len)
{
	CHECK_LOCK_STATE_RETURN_NEG
		
	LOCK
	
	int toRet;
#ifdef CLOG_SAFE_MODE
	s_out = fopen("out.log", "a");

	if (s_out > 0)
	{
 #ifdef CLOG_ENABLE_TIMESTAMPS
		SysTimestamp(s_out);
 #endif
		fwrite("[Debug]: ", sizeof(char), 9, s_out);
		toRet = fwrite(aMsg, sizeof(char), len, s_out);
		fwrite("\n", sizeof(char), 1, s_out);
		fclose(s_out);
		s_out = NULL;
		UNLOCK
		return toRet;
	}
#else
	if (s_out > 0)
	{
 #ifdef CLOG_ENABLE_TIMESTAMPS
		SysTimestamp(s_out);
 #endif
		
		fwrite("[Debug]: ", sizeof(char), 9, s_out);
		toRet = fwrite(aMsg, sizeof(char), len, s_out);
		fwrite("\n", sizeof(char), 1, s_out);
		UNLOCK
		return toRet;
	}
#endif

	UNLOCK
	return -1;
}

int CLog::Error(const char* aErr, ...)
{
	CHECK_LOCK_STATE_RETURN_NEG
		
	LOCK
	
#ifdef CLOG_SAFE_MODE
	s_err = fopen("err.log", "a");
#endif

	if (s_err > 0)
	{
		int toRet;
		va_list ap;
		
#ifdef CLOG_ENABLE_TIMESTAMPS
		SysTimestamp(s_err);
#endif
		va_start(ap, aErr);
		toRet = vfprintf(s_err, aErr, ap);
		va_end(ap);
		fflush(s_err);
#ifdef CLOG_SAFE_MODE
		fclose(s_err);
		s_err = NULL;
#endif
		UNLOCK
		return toRet;
	}
	
	UNLOCK
	return -1;
}

int CLog::ErrorRaw(const char* aErr, int len)
{
	CHECK_LOCK_STATE_RETURN_NEG
	
	LOCK

	int toRet;
	
#ifdef CLOG_SAFE_MODE
	s_err = fopen("err.log", "a");
	if (s_err > 0)
	{
 #ifdef CLOG_ENABLE_TIMESTAMPS
		SysTimestamp(s_err);
 #endif
		toRet = fwrite(aErr, sizeof(char), len, s_err);
		fclose(s_err);
		s_err = NULL;
		UNLOCK
		return toRet;
	}
#else
	if (s_err > 0)
	{
 #ifdef CLOG_ENABLE_TIMESTAMPS
		SysTimestamp(s_err);
 #endif
		toRet = fwrite(aErr, sizeof(char), len, s_err);
		UNLOCK
		return toRet;
	}
#endif
	
	UNLOCK
	return -1;
}

int CLog::Critical(const char* aErr, ...)
{
	CHECK_LOCK_STATE_RETURN_NEG
	
	LOCK
	
#ifdef CLOG_SAFE_MODE
    s_err = fopen("err.log", "a");
#endif

	if (s_err > 0)
	{
		int toRet;
		va_list ap;

#ifdef CLOG_ENABLE_TIMESTAMPS
		SysTimestamp(s_err);
#endif
		va_start(ap, aErr);
		fwrite("[Critical]: ", sizeof(char), 12, s_err);
		toRet = vfprintf(s_err, aErr, ap);
		fwrite("\n", sizeof(char), 1, s_err);
		va_end(ap);
		fflush(s_err);
#ifdef CLOG_SAFE_MODE
        fclose(s_err);
        s_err = NULL;
#endif
		UNLOCK
		return toRet;
	}
	
	UNLOCK
	return -1;
}

int CLog::CriticalRaw(const char* aErr, int len)
{
	CHECK_LOCK_STATE_RETURN_NEG
	
	LOCK
	
	int toRet;
#ifdef CLOG_SAFE_MODE
	s_err = fopen("err.log", "a");
	if (s_err > 0)
	{
 #ifdef CLOG_ENABLE_TIMESTAMPS
		SysTimestamp(s_err);
 #endif
		fwrite("[Critical]: ", sizeof(char), 12, s_err);
		toRet = fwrite(aErr, sizeof(char), len, s_err);
		fwrite("\n", sizeof(char), 1, s_out);
		fclose(s_err);
		s_err = NULL;
		UNLOCK
		return toRet;
	}
#else
	if (s_err > 0)
	{
 #ifdef CLOG_ENABLE_TIMESTAMPS
		SysTimestamp(s_err);
 #endif
		fwrite("[Critical]: ", sizeof(char), 12, s_err);
		toRet = fwrite(aErr, sizeof(char), len, s_err);
		fwrite("\n", sizeof(char), 1, s_out);
		UNLOCK
		return toRet;
	}
#endif
	
	UNLOCK
	return -1;
}

int CLog::SysOpen(int aOverwrite)
{
#ifdef CLOG_SAFE_MODE
	if (aOverwrite)
	{
		s_out = fopen("out.log", "w");
		s_err = fopen("err.log", "w");
		fclose(s_out);
		fclose(s_err);
	}

	return 1;
#else
	// Only allow Open() to be called once! If not, could end up overwriting the
	// log files
	if (SysIsOpen())
	{
		return 0;
	}
	
	// SysIsOpen will only return true if both files are open, we must ensure that
	// both are closed if we're going to continue
	SysClose();
	
	if (aOverwrite)
	{
		s_out = fopen("out.log", "w");
		s_err = fopen("err.log", "w");
	}
	else
	{
		s_out = fopen("out.log", "a");
		s_err = fopen("err.log", "a");
	}
	
	return (s_out > 0 && s_err > 0);
#endif
}

void CLog::SysClose()
{
	if (s_out > 0)
	{
		fflush(s_out);
		fclose(s_out);
		s_out = NULL;
	}
	
	if (s_err > 0)
	{
		fflush(s_err);
		fclose(s_err);
		s_err = NULL;
	}
}

int CLog::SysIsOpen()
{
#ifdef CLOG_SAFE_MODE
	// The file is never open in safe mode other than in the write methods
	return 0;
#else
	return (s_out > 0 && s_err > 0);
#endif
}

void CLog::SysTimestamp(FILE* f)
{
	char timeStr[16];
	time_t tm = time(NULL);
	size_t written;
	
	written = strftime(timeStr, 16, "[%H:%M:%S] ", localtime(&tm));
	fwrite(timeStr, sizeof(char), written, f);
}

void CLog::lock()
{
#ifdef CLOG_ENABLE_LOCKS
	sceKernelWaitSema(m_sema, 1, NULL);
#endif
}

void CLog::unlock()
{
#ifdef CLOG_ENABLE_LOCKS
	sceKernelSignalSema(m_sema, 1);
#endif
}
