/*******************************************************************************

  PSPoste - Wireless Email Client for the Sony PSP
  Copyright (C) Benjamin Biron, Sajeeth Cherian, David Cummings

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

  Module :	CMailbox.cpp

  Description :	

  Last Modified $Date: 2008/10/01 03:03:21 $

  $Revision: 1.11 $

*******************************************************************************/

//******************************************************************************
//   Include Files
//******************************************************************************
#include "CMailbox.h"
#include "CPOP3Retriever.h"
#include "CSMTPSender.h"
#include "CMaildirStore.h"
#include "CFileSystem.h"
#include "CConnectionManager.h"
#include "CLog.h"
#include "CSmtp.h"
#include "CCertRequest.h"
#include "main_globals.h"

//******************************************************************************
//   Local Macros
//******************************************************************************

//******************************************************************************
//   Local Constants
//******************************************************************************

//******************************************************************************
//   Static Prototypes
//******************************************************************************

//******************************************************************************
//   Global Variables
//******************************************************************************

//******************************************************************************
//   Static Variables
//******************************************************************************

//******************************************************************************
//   Class Definition
//******************************************************************************

//******************************************************************************
//	 This function is used in the sorting of the message headers. It is required
//	 by the STL and will simply call the call the compare on the mailbox so it
//	 can be sorted according to the mailbox setting.
//******************************************************************************
bool messageHeaderBinaryPredicate(CMessageHeader* aHeader1,
                                  CMessageHeader* aHeader2)
{
	return (aHeader1->m_mailbox->compareMessageHeaders(aHeader1, aHeader2) < 0);
}

//******************************************************************************
//	
//******************************************************************************
CMailbox::CMailbox(CAccount* account) : m_pReadyToSend(false),
										m_send_mail(false),
										m_sending(false),
										m_unreadMessageCount(0),
                                        m_account(account),
                                        m_sort(new CMailboxSortDate),
                                        m_check_mail(false),
                                        m_waiting_for_con_man(false),
                                        m_retrieving(false),
                                        m_last_retrieved_num(0)
{
	// Create the cert manager & request handler
	CString certPath = CFileSystem::MakeFullPath("Data/certs");
	if (!CFileSystem::DirectoryExists(certPath))
	{
		if (!CFileSystem::MakeDirectory(certPath))
		{
			CLog::Error("Error creating directory: %s\n", certPath.GetPtr());
		}
	}
	
	m_certManager = vmime::create<CCertificateManager>(certPath, CFileSystem::MakeFullPath("Data/flash_certs"));
	m_certRequestHandler = new CCertRequestHandler(m_certManager);
	
	// Create the Mbx's for communicating with the retriever and sender threads
	CString mbxName = account->display_name;
	m_mailbox_mbx = sceKernelCreateMbx(mbxName + "_Mailbox", 0, NULL);
	m_retriever_mbx = sceKernelCreateMbx(mbxName + "_Retriever", 0, NULL);
	m_sender_mbx = sceKernelCreateMbx(mbxName + "_Sender", 0, NULL);
	
	// Create the retriever, sender and store
	m_store = new CMaildirStore(getMaildirRelativePath(), this, m_mailbox_mbx);
	m_retriever = new CPOP3Retriever(m_store, m_account, m_mailbox_mbx, m_retriever_mbx, m_certManager);
	m_sender = new CSMTPSender(m_account, m_mailbox_mbx, m_sender_mbx, m_certManager);

	m_retriever_thread = sceKernelCreateThread(mbxName + " Retriever_Thread",
	                                           retrieverThreadCB, 0x11,
	                                           RETRIEVER_THREAD_STACK_SIZE,
	                                           PSP_THREAD_ATTR_USER, NULL);
	if (m_retriever_thread & 0x80000000)
	{
		CLog::Error("Failed to create the retriver thread: 0x%x\n",
		            m_retriever_thread);
	}
	else
	{
		sceKernelStartThread(m_retriever_thread, sizeof(CPOP3Retriever),
		                     reinterpret_cast<void*>(m_retriever));	
	}
	
	m_sender_thread = sceKernelCreateThread(mbxName + " Sender_Thread",
	                                        senderThreadCB, 0x11,
	                                        SENDER_THREAD_STACK_SIZE,
	                                        PSP_THREAD_ATTR_USER, NULL);

	if (m_sender_thread & 0x80000000)
	{
		CLog::Error("Failed to create the sender thread: 0x%x\n",
		            m_sender_thread);
	}
	else
	{
		sceKernelStartThread(m_sender_thread, sizeof(CSMTPSender),
		                     reinterpret_cast<void*>(m_sender));
	}

	// Create the stored message list and populate
	m_pMessageList = new CMailbox::TMessageList();
	m_store->getMessages(m_pMessageList);
	
	// We must update the m_last_retrieved_num so that the first time we
	// retrieve messages, we don't insert the already retrieved messages. Note
	// that forgetting this is what caused the "duplication bug".
	m_last_retrieved_num = m_store->getMessageCount();

	// Create send message list
	m_pSendMsgList = new CMailbox::TSendMsgList();

	// Get the unread count
	CMailbox::TMessageList::iterator it = m_pMessageList->begin();
	while (it != m_pMessageList->end())
	{
		if ((*it)->unread)
			m_unreadMessageCount++;
		it++;
	}
}

CMailbox::~CMailbox() {
	clearMsgList();
	clearSendMsgList();
	clearSendMsgList(m_pReadyToSend);

    SAFE_DELETE(m_pMessageList);
	SAFE_DELETE(m_pSendMsgList);
	SAFE_DELETE(m_pReadyToSend);

	sceKernelCancelReceiveMbx(m_mailbox_mbx, NULL);
	sceKernelCancelReceiveMbx(m_retriever_mbx, NULL);
	sceKernelDeleteMbx(m_mailbox_mbx);
	sceKernelDeleteMbx(m_retriever_mbx);
	
	sceKernelTerminateDeleteThread(m_retriever_thread);
	sceKernelTerminateDeleteThread(m_sender_thread);
	
	SAFE_DELETE(m_retriever);
	SAFE_DELETE(m_store);
	SAFE_DELETE(m_sender);
	
	SAFE_DELETE(m_certRequestHandler);
	SAFE_DELETE(m_sort);
}

void CMailbox::clearMsgList()
{
	TMessageList::iterator it = m_pMessageList->begin();

	while( it !=  m_pMessageList->end() ) {
		SAFE_DELETE((*it));
		it++;
	}

	m_pMessageList->clear();
}

void CMailbox::clearSendMsgList(TSendMsgList* aSendMsgList)
{
	// Note that this function takes in a paramater which corresponds to the
	// list to clear, if the value is null, it will use m_pSendMsgList.
	//
	// Also note that it does not delete the list passed in!
	if (aSendMsgList == NULL)
		aSendMsgList = m_pSendMsgList;
	
	if (aSendMsgList == NULL)
		return;

	TSendMsgList::iterator it = aSendMsgList->begin();

	while( it != aSendMsgList->end() ) {
		// SAFE_DELETE is pointless here since the pointers will get lossed once
		// the list is cleared
		delete *it;
		it++;
	}

	m_pSendMsgList->clear();
}

void CMailbox::Process()
{
	// Note that the message must be static because the mbx will just pass the
	// pointer. This is safe only because the retreive message doesn't have any
	// variable data, if it did, we'd have to use the heap.
	static SRetrieverMsg retrieverMsg;
	SMailboxMsg* mailboxMsg;
	int err;
	
	// Process any pending certificate request
	m_certRequestHandler->processCertRequests();
	
	// Poll the mbx
	if (sceKernelPollMbx(m_mailbox_mbx, (void**)&mailboxMsg) == 0)
	{
		// A mbx message has been retrieved
		switch (mailboxMsg->m_type)
		{
			case EMailboxMsgType_RetrieveDisconnected:
				m_retrieving = false;
				if (!m_sending)
					CConnectionManager::RequestDisconnect();
				break;
			case EMailboxMsgType_SendDisconnected:
				m_sending = false;
				if (!m_retrieving)
					CConnectionManager::RequestDisconnect();
				break;
			case EMailboxMsgType_StoringDone:
				insertNewMessages();
				m_unreadMessageCount += mailboxMsg->count;
				notifyEvent(mailboxMsg);
				break;
			case EMailboxMsgType_SendFailed:
				// TODO: Use the failed list to update my messages to send list
				// Don't need this for current functionality
				clearSendMsgList(mailboxMsg->sentList);
				SAFE_DELETE(mailboxMsg->sentList);
				clearSendMsgList(mailboxMsg->failedList);
				SAFE_DELETE(mailboxMsg->failedList);
				notifyEvent(mailboxMsg);
				break;
			case EMailboxMsgType_SendDone:
				clearSendMsgList(mailboxMsg->sentList);
				SAFE_DELETE(mailboxMsg->sentList);
				notifyEvent(mailboxMsg);
				break;
			case EMailboxMsgType_CertRequest:
				m_certRequestHandler->incomingCertRequest(mailboxMsg->certRequest);
				// There's no "incoming cert" user event, so no need to call
				// notifyEvent
				break;
			default:
				notifyEvent(mailboxMsg);
				break;
		}
		
		SAFE_DELETE(mailboxMsg);
	}
	
	// This state machine runs from the moment a "check mail" request is
	// received to the moment the "retrieve" message has been sent to the
	// retriever or an error bringing up the connection has been detected.
	//
	// It is basically responsible for bringing up the WiFi connection, sending
	// the "retrieve" message to the retriever, or notifying the listeners of
	// connection errors.
	//
	// Note that connection errors and actual retrieval errors are both called
	// "retrieval errors".
	if (m_check_mail)
	{
		switch (CConnectionManager::GetState())
		{
			case EConnManState_Connecting:
				break;
			case EConnManState_Connected:
				m_retrieving = true;
				if ((err = sceKernelSendMbx(m_retriever_mbx, &retrieverMsg)))
					CLog::Error("Error sending to retriever mbx: %x\n", err);
					
				m_check_mail = false;
				m_waiting_for_con_man = false;
				break;
			case EConnManState_Disconnected:
				if (!m_waiting_for_con_man)
				{
					CConnectionManager::ClearError();
				    if (CConfigFile::s_pSettings->connection_selected != 0)
				        CConnectionManager::RequestConnect(CConfigFile::s_pSettings->connection_selected);
				    else
				        CConnectionManager::RequestConnect();

					m_waiting_for_con_man = true;
				}
				else
				{
					if (int conErr = CConnectionManager::CheckError())
					{
						SMailboxMsg failedMsg(EMailboxMsgType_RetrieveFailed);
						failedMsg.errorText = CConnectionManager::GetErrorText(conErr);
						notifyEvent(&failedMsg);
						m_check_mail = false;
						m_waiting_for_con_man = false;
						return;
					}
				}
				break;
			default:
				CLog::Error("Unknown connection manager state\n");
				break;
		}
	}

	// This state machine runs from the moment a "send mail" request is
	// received to the moment the ESenderMsgType_Send message has been sent to the
	// sender or an error bringing up the connection has been detected.
	//
	// It is basically responsible for bringing up the WiFi connection, sending
	// the "send" message to the retriever, or notifying the listeners of
	// connection errors.
	//
	// Note that connection errors and actual send errors are both called
	// "send errors".
	if (m_send_mail)
	{
		SSenderMsg* senderMsg = NULL;
		switch (CConnectionManager::GetState())
		{
			case EConnManState_Connecting:
				break;
			case EConnManState_Connected:
				m_sending = true;	

				senderMsg = new SSenderMsg(ESenderMsgType_Send);
				senderMsg->m_msgList = m_pReadyToSend;
				
				if ((err = sceKernelSendMbx(m_sender_mbx, senderMsg))) {
					CLog::Error("Error sending to sender mbx: %x\n", err);
					// TODO: put items ready to send back in the main list
					// Don't need this for current functionality
				}	
				m_pReadyToSend = NULL;

				m_send_mail = false;
				m_waiting_for_con_man = false;
				break;
			case EConnManState_Disconnected:
				if (!m_waiting_for_con_man)
				{
					CConnectionManager::ClearError();
				    if (CConfigFile::s_pSettings->connection_selected != 0)
				        CConnectionManager::RequestConnect(CConfigFile::s_pSettings->connection_selected);
				    else
				        CConnectionManager::RequestConnect();

					m_waiting_for_con_man = true;
				}
				else
				{
					if (int conErr = CConnectionManager::CheckError())
					{
						// TODO: !!! Need to include the items in m_pReadyToSend in this msg
						m_pReadyToSend = NULL;
						SMailboxMsg failedMsg(EMailboxMsgType_SendFailed);
						failedMsg.errorText = CConnectionManager::GetErrorText(conErr);
						notifyEvent(&failedMsg);
						m_send_mail = false;
						m_waiting_for_con_man = false;
						return;
					}
				}
				break;
			default:
				CLog::Error("Unknown connection manager state\n");
				break;
		}
	}
}

bool CMailbox::ProcessInput()
{
	return false;
}
		
void CMailbox::subscribe(MMailboxEventListener* aListener)
{
	m_mailbox_listeners.push_back(aListener);
}

void CMailbox::unsubscribe(MMailboxEventListener* aListener)
{
	m_mailbox_listeners.remove(aListener);
}

int CMailbox::deleteMessage(CMessageHeader* aMsg)
{
	// Messages will first get deleted locally, then on the server if the user
	// has the account configured for remote deletion. The remote deletion takes
	// place in the retriever, much like the retrieval, this is asynchronous and
	// listeners will be notified. Listeners will be notified immediately if
	// remote deletion is not selected.
	int err = 0;
	
	// Flag them as deleted in the store
	if ((err = m_store->deleteMessage(aMsg->num)))
		return err;

	// Delete from the list of headers
	TMessageList::iterator it = m_pMessageList->begin();
	while (it != m_pMessageList->end())
	{
		if ((*it) == aMsg)
		{
			if ((*it)->unread)
				m_unreadMessageCount--;
				
			SAFE_DELETE(*it);
			m_pMessageList->erase(it);
			break;
		}
		it++;
	}
	
	// BEN TODO: Delete from server?
	return err;
}

int CMailbox::deleteMarkedMessages()
{
	int err = 0;
	int unreadCount = 0;
	std::vector<int> numsToDel;
	std::vector<TMessageList::iterator> msgsToDel;
	TMessageList::iterator it = m_pMessageList->begin();
	
	// Find the marked messages
	while (it != m_pMessageList->end())
	{
		if ((*it)->marked)
		{
			numsToDel.push_back((*it)->num);
			msgsToDel.push_back(it);
			if ((*it)->unread)
				unreadCount++;
		}
		it++;
	}
	
	// Flag them as deleted in the store
	if ((err = m_store->deleteMessages(numsToDel)))
		return err;
		
	m_unreadMessageCount -= unreadCount;
		
	// Delete them from the list of headers
	// Note that we must traverse the list in reverse order so when we delete
	// an item from the list, the remaining items in the list don't have to be
	// adjusted to compensate for the removed item.
	std::vector<TMessageList::iterator>::iterator revIt = msgsToDel.end();
	while (revIt != msgsToDel.begin())
	{
		revIt--;
		SAFE_DELETE(*(*revIt));
		m_pMessageList->erase(*revIt);
	}
	
	return err;
}

int CMailbox::retrieveNewMessages()
{
	if (isRetrieving())
		return -1;

	m_check_mail = true;
	
	return 0;
}

void CMailbox::cancelRetrieve()
{
	// BEN TODO: Implement
}

void CMailbox::requestBodyPart(u32 aId)
{
	// BEN TODO: Implement
}

void CMailbox::getBodyPart(u32 aId)
{
	// BEN TODO: Implement
}
	
int CMailbox::getUnreadMessageCount()
{
	return m_unreadMessageCount;
}

int CMailbox::getReadMessageCount()
{
	return getTotalMessageCount() - getUnreadMessageCount();
}

int CMailbox::getTotalMessageCount()
{
	return m_pMessageList->size();
}

bool CMailbox::isRetrieving()
{
	return m_check_mail || m_retrieving;
}

int CMailbox::compareMessageHeaders(const CMessageHeader* aHeader1,
		                  const CMessageHeader* aHeader2) const
{
	return m_sort->compare(aHeader1, aHeader2);
}

CString CMailbox::getMaildirRelativePath()
{
	// BEN TODO: Normalize the file paths... either always absolute or always
	// relative (both psplink and release modes). Should the drive be included?
	//
	// Part of the problem is because the vmime pspFile.cpp file uses newlib for
	// file/directory handling while CFileSystem.cpp uses the sce io code, and
	// they both have their differences in path representation. Ex: sceIoMkdir
	// fails if the path ends with a trailing "/". And the newlib calls fail if
	// the ms0:/ prefix is prepended to the path.
	CString fullPath(CFileSystem::MakeFullPath(""));
	
	fullPath += "store";
	if (!CFileSystem::DirectoryExists(fullPath))
	{
		if (!CFileSystem::MakeDirectory(fullPath))
		{
			CLog::Error("Error creating directory: %s\n", fullPath.GetPtr());
		}
	}
		
	fullPath += "/";
	fullPath += m_account->display_name;
	if (!CFileSystem::DirectoryExists(fullPath))
	{
		if (!CFileSystem::MakeDirectory(fullPath))
		{
			CLog::Error("Error creating directory: %s\n", fullPath.GetPtr());
		}
	}

	CString toRet = "/store/";
	toRet += m_account->display_name;
	toRet += "/"; // Yes, this is required.
	return toRet;
}

CString CMailbox::getMaildirAbsolutePath()
{
	// BEN NOTES:
	//  - MakeFullPath("") returns the root's absolute path
	//  - MakeDirectory() does not handle trailing slashes at all!!
	//  - SplitPath() needs a trailing slash to return the full path when no
	//    filename is provided
	// BEN TODO: Normalize the file paths... either always absolute or always
	// relative (both psplink and release modes). Should the drive be included?
	CString toRet;
	CString fullPath(CFileSystem::MakeFullPath(""));
	
	fullPath += getMaildirRelativePath();
	CFileSystem::SplitPath(fullPath, NULL, &toRet, NULL, NULL);
	return toRet;
}

// This function retrieves all messages retrieved since last time it was called.
// It is responsible for updating m_last_retrieved_num.
void CMailbox::insertNewMessages()
{
	TMessageList* newMessages = new TMessageList;
	m_store->getMessages(newMessages, false, m_last_retrieved_num+1);
	
	TMessageList::iterator it = newMessages->begin();
	while (it != newMessages->end())
	{
		bisectionInsert(*it);
		it++;
	}
	
	m_last_retrieved_num = (*newMessages)[newMessages->size()-1]->num;
	delete newMessages;
}

// The bisectionInsert will insert the header in the order determined by the
// mailbox sort
void CMailbox::bisectionInsert(CMessageHeader* aHeader)
{
	TMessageList::iterator iBegin = m_pMessageList->begin();
	TMessageList::iterator iEnd = m_pMessageList->end();
	TMessageList::iterator iSorter = iBegin;
	int steps = m_pMessageList->size();

	// Find a place to put the contact
	while (iBegin != iEnd)
	{
	  // start with the iterator at the current beginning of the list
	  iSorter = iBegin;
	  // find the middle
	  steps /= 2;
	  // move the iterator to the middle
	  iSorter += steps;
	  
	  // use the > of CString to see if the name is greater than the iterator name
	  // if so move the beginning up to the new middle and restart
	  // 
	  if (aHeader->m_mailbox->compareMessageHeaders(aHeader, *iSorter)  > 0)
	  {
		 // change the beginning of the list to the current  
		 // iterator position
		 iBegin = iSorter;
		 // if we didn't move at all, and if we aren't at the  
		 // end of the list,  move the beginning one more step.
		 if (steps == 0 && iBegin != iEnd)
		 {
			++iBegin;
		 }
		 // we need to do it this way because eventually, you just 
		 // run out of "steps" (it's equal to 0), and the routine 
		 // would just sit there on the same iterator forever.  If 
		 // it gets to this point, it's safe to assume that simply 
		 // moving the iterator one more step in the appropriate 
		 // direction will locate the correct insertion point.
	  }
	  else
	  {
		 iEnd = iSorter;
		 if (steps == 0 && iEnd != iBegin)
		 {
			--iEnd;
		 }
	  }
	}
	
	iSorter = iEnd;
	m_pMessageList->insert(iSorter, aHeader);
}

void CMailbox::notifyEvent(SMailboxMsg* aEvent)
{
	std::list<MMailboxEventListener*>::iterator it = m_mailbox_listeners.begin();
	
	// Shouldn't be too many listeners so a switch inside the loop shouldn't be
	// too bad
	while (it != m_mailbox_listeners.end())
	{
		switch (aEvent->m_type)
		{
			case EMailboxMsgType_RetrievalStarted:
				(*it)->retrievalStarted(this, aEvent->count);
				break;
			case EMailboxMsgType_RetrievedOne:
				(*it)->retrievedOne(this, aEvent->index, aEvent->count);
				break;
			case EMailboxMsgType_RetrievalDone:
				(*it)->retrievalDone(this, aEvent->count);
				break;
			case EMailboxMsgType_RetrievedNone:
				(*it)->retrievedNone(this);
				break;
			case EMailboxMsgType_RetrieveFailed:
				(*it)->retrieveFailed(this, aEvent->errorText);
				break;
			case EMailboxMsgType_SentOne:
				(*it)->sentOne(this, aEvent->index, aEvent->count);
				break;
			case EMailboxMsgType_SendDone:
				(*it)->sendDone(this, aEvent->count);
				break;
			case EMailboxMsgType_SendFailed:
				(*it)->sendFailed(this, aEvent->errorText);
				break;
			case EMailboxMsgType_StoringStarted:
				(*it)->storingStarted(this, aEvent->count);
				break;
			case EMailboxMsgType_StoredOne:
				(*it)->storedOne(this, aEvent->index, aEvent->count);
				break;
			case EMailboxMsgType_StoringDone:
				(*it)->storingDone(this, aEvent->count);
				break;
			case EMailboxMsgType_StoredNone:
				(*it)->storedNone(this);
				break;
			case EMailboxMsgType_StoreFailed:
				(*it)->storeFailed(this, aEvent->errorText);
				break;
			default:
				CLog::Error("Unknown message (%d)\n", aEvent->m_type);
				break;
		}

		it++;
	}
}

CString CMailbox::getBody(CMessageHeader* aMsg)
{
	return m_store->getBody(aMsg);
}

void CMailbox::markRead(CMessageHeader* aMsg, bool aIsRead)
{
	m_store->markRead(aMsg, aIsRead);

	if (aMsg->unread == aIsRead)
	{
		aMsg->unread = !aIsRead;
		aIsRead ? m_unreadMessageCount-- : m_unreadMessageCount++;
	}
	
}

void CMailbox::setSort(MMailboxSort *aSort)
{
	SAFE_DELETE(m_sort);
	m_sort = aSort;
	m_store->sortMessageList(m_pMessageList);
}

void CMailbox::addSendMsg(CSendMsg* msg)
{
	m_pSendMsgList->push_back(msg);
}

int CMailbox::sendAllMessages() 
{
	if (m_pSendMsgList->size() == 0) {
		return SEND_ERROR_NO_MESSAGES;
	}

	if (isSending()) {
		return SEND_ERROR_ALREADY_SENDING;
	}

	// Build the ready to send list
	m_pReadyToSend = new TSendMsgList();

	CMailbox::TSendMsgList::iterator it = m_pSendMsgList->begin();
	while(it != m_pSendMsgList->end()) 
	{
		m_pReadyToSend->push_back(*it);
		it++;
	}

	// Clear the send message list
	m_pSendMsgList->clear();

	m_send_mail = true;
	return 0;
}

bool CMailbox::isSending()
{
	return m_send_mail || m_sending;
}
//*******************************  END OF FILE  ********************************
