/*******************************************************************************

  PSPoste - Wireless Email Client for the Sony PSP
  Copyright (C) Benjamin Biron, Sajeeth Cherian, David Cummings

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

  Module :	CPOP3Retriever.cpp

  Description :	

*******************************************************************************/

#include "CPOP3Retriever.h"
#include "CMaildirStore.h"
#include "CConnectionManager.h"
#include "CLog.h"
#include "CAssert.h"
#include "CMailbox.h"

CPOP3Retriever::CPOP3Retriever(CMaildirStore* aLocalStore, CAccount* aAccount, SceUID aMailboxMbx,
                               SceUID aRetrieverMbx, vmime::ref<CCertificateManager> aCertManager) :
	CMailRetriever(aMailboxMbx, aRetrieverMbx, aCertManager),
	m_account(*aAccount),
	m_mailbox_mbx(aMailboxMbx),
	m_retriever_mbx(aRetrieverMbx),
	m_local_store(aLocalStore),
	m_last_retrieved(0)
{
	CString pop3Url;
	if (m_account.incoming_server.Substring(0, 7).IEquals("pop3://") || m_account.incoming_server.Substring(0, 7).IEquals("pop3s://"))
	{
		pop3Url = m_account.incoming_server;
	}
	else
	{
		if (m_account.incoming_server_tls)
			pop3Url = "pop3s://";
		else
			pop3Url = "pop3://";
			
		pop3Url += m_account.incoming_server;
		if (m_account.incoming_server_port != 0)
		{
			pop3Url += ":";
			pop3Url += m_account.incoming_server_port;
		}
	}
	
	try
	{
		VUrl url(pop3Url);
	
		// Create a new session
		m_sess_ref = vmime::create <VSession>();
		
		// Create an instance of the store
		m_store_ref = m_sess_ref->getStore(url);
		m_store_ref->setCertificateVerifier(m_certManager);
		
		if (m_account.incoming_server_username != "" || m_account.incoming_server_password != "")
		{
			m_store_ref->setProperty("auth.username", m_account.incoming_server_username);
			m_store_ref->setProperty("auth.password", m_account.incoming_server_password);
		}
	}
	catch (vmime::exception& e)
	{
		V_EXCEPT(e);
	}
	catch (std::exception& e)
	{
		S_EXCEPT(e);
	}
}

CPOP3Retriever::~CPOP3Retriever()
{
}

void CPOP3Retriever::retrieve()
{
	try
	{
		connect();
		open();

		// Fetch all the messages
		//CString out;
		//out.Printf("Last Retrieved = %d", m_last_retrieved);
		//ASSERT(0, out.GetPtr());
		VMessageRefList serverMessages = m_folder_ref->getMessages(m_last_retrieved + 1);
		
		// First get the UIDs of the messages to determine the ones that don't
		// exist in the local store
		m_folder_ref->fetchMessages(serverMessages, vmime::net::folder::FETCH_UID);
		
		// Now filter out the ones that already exist in the store. Prune the
		// list only if the list of messages corresponds to the list of all
		// messages on the server.
		m_local_store->filterAlreadyRetrievedMessages(serverMessages, m_last_retrieved == 0);
		
		if (serverMessages.size() > 0)
		{
			// Actually fetch the message contents
			m_folder_ref->fetchMessages(serverMessages,
			                            m_folder_ref->getFetchCapabilities(),
			                            this);
			/*
			m_folder_ref->fetchMessages(serverMessages,
			                            vmime::net::folder::FETCH_FULL_HEADER |
			                            vmime::net::folder::FETCH_UID |
			                            vmime::net::folder::FETCH_SIZE |
			                            vmime::net::folder::FETCH_IMPORTANCE |
			                            vmime::net::folder::FETCH_CONTENT_INFO,
			                            this);
			*/
			
			// Add the messages to the store
			m_local_store->addMessages(serverMessages);
			
			// Now that everything has been added, update m_last_retrieved
			m_last_retrieved = serverMessages[serverMessages.size()-1]->getNumber();
		}
		CLog::Message("Retrieved %d messages from pop3://%s\n", serverMessages.size(),
		                                                        m_account.incoming_server.GetPtr());
		
		if (serverMessages.size() == 0)
		{
			stop(0); // Will send out a RetrievedNone msg
		}
	}
	catch (vmime::exceptions::message_not_found& e)
	{
		// We can safely ignore this as all it means is that no new messages
		// have been retrieved.		
		CLog::Debug("No new mail on server");
		stop(0); // Will send out a RetrievedNone msg
	}
	catch (CCertificateException& e)
	{
		// In this case, we don't want to send a Disconnected message 'cause
		// it'd bring down the wifi connection. We also don't need to close or
		// disconnect from the store since we never connected.
		m_certComm.sendCertRequest(e.cert());
		return;
	}
	catch (vmime::exception& e)
	{
		V_EXCEPT(e);
		failed(e.what());
	}
	catch (std::exception& e)
	{
		S_EXCEPT(e);
		failed(e.what());
	}
	
	try
	{
		close();
		disconnect();
	}
	catch (vmime::exception& e)
	{
		V_EXCEPT(e);
	}
	catch (std::exception& e)
	{
		S_EXCEPT(e);
	}
	
	disconnected(); // Sends a Disconnected message
}

void CPOP3Retriever::certDeclined()
{
	// Let's send a message back to the mailbox telling it to tear down the wifi
	// connection
	failed("Certificate declined.");
	disconnected();
}

void CPOP3Retriever::certAccepted()
{
	retrieve();
}

void CPOP3Retriever::connect()
{
	if (!m_store_ref)
		throw vmime::exception("Store was uninitialized");
		
	if (!m_store_ref->isConnected())
		m_store_ref->connect();
}

void CPOP3Retriever::disconnect()
{
	if (!m_store_ref)
		throw vmime::exception("Store was uninitialized");
		
	if (m_store_ref->isConnected())
		m_store_ref->disconnect();
}

void CPOP3Retriever::open()
{
	if (m_folder_ref && m_folder_ref->isOpen())
		return;
		
	m_folder_ref = m_store_ref->getDefaultFolder();
	if (!m_folder_ref->isOpen())
		m_folder_ref->open(V_MODE_READ_ONLY);
}

void CPOP3Retriever::close()
{	
	if (m_folder_ref && m_folder_ref->isOpen())
	{
		// BEN NOTE: POP3 Folders automatically expunge at end of session
		m_folder_ref->close(false);
		m_folder_ref = NULL;
	}
}

/* Progress listener functions follow */

/* When true is returned, will cancel the retrieval; not supported by POP3 */
const bool CPOP3Retriever::cancel() const
{
	return false;
}

/* Notifies the progress listener that a retrieval has started */
void CPOP3Retriever::start(const int predictedTotal)
{
	int err;
	
	SMailboxMsg* startMsg = new SMailboxMsg(EMailboxMsgType_RetrievalStarted);
	startMsg->count = predictedTotal;
	
	if ((err = sceKernelSendMbx(m_mailbox_mbx, startMsg)) < 0)
	{
		CLog::Error("sceKernelSendMbx returned %x\n", err);
	}
}

/* `current` has been retrieved */
void CPOP3Retriever::progress(const int current, const int currentTotal)
{
	int err;
	
	SMailboxMsg* progressMsg = new SMailboxMsg(EMailboxMsgType_RetrievedOne);
	progressMsg->index = current;
	progressMsg->count = currentTotal;
	
	if ((err = sceKernelSendMbx(m_mailbox_mbx, progressMsg)) < 0)
	{
		CLog::Error("sceKernelSendMbx returned %x\n", err);
	}
}

/* Retrieval done */
void CPOP3Retriever::stop(const int total)
{
	int err;
	
	SMailboxMsg* stopMsg;
	
	if (total == 0)
		stopMsg = new SMailboxMsg(EMailboxMsgType_RetrievedNone);
	else
		stopMsg = new SMailboxMsg(EMailboxMsgType_RetrievalDone);
	
	stopMsg->count = total;
	
	if ((err = sceKernelSendMbx(m_mailbox_mbx, stopMsg)) < 0)
	{
		CLog::Error("sceKernelSendMbx returned %x\n", err);
	}
}

void CPOP3Retriever::failed(CString aErrorText)
{
	int err;
	
	SMailboxMsg* failedMsg = new SMailboxMsg(EMailboxMsgType_RetrieveFailed);
	failedMsg->errorText = aErrorText;
	
	if ((err = sceKernelSendMbx(m_mailbox_mbx, failedMsg)) < 0)
	{
		CLog::Error("sceKernelSendMbx returned %x\n", err);
	}
}

void CPOP3Retriever::disconnected()
{
	int err;
	
	SMailboxMsg* disconnectedMsg = new SMailboxMsg(EMailboxMsgType_RetrieveDisconnected);
	
	if ((err = sceKernelSendMbx(m_mailbox_mbx, disconnectedMsg)) < 0)
	{
		CLog::Error("sceKernelSendMbx returned %x\n", err);
	}
}
