/*******************************************************************************

  PSPoste - Wireless Email Client for the Sony PSP
  Copyright (C) T Swann, Benjamin Biron, Sajeeth Cherian, David Cummings

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

  Module :	CTabbedWindow.cpp

  Description :	

  Last Modified $Date: 2007/02/05 01:28:40 $

  $Revision: 1.6 $

*******************************************************************************/

//******************************************************************************
//   Include Files
//******************************************************************************
#include "CTabbedWindow.h"
#include "CGfx.h"
#include "CFrameWork.h"
#include "CFont.h"
#include "CWindow.h"
#include "CInput.h"
#include "CSkinManager.h"
#include "irda_KeyboardLayouts.h"
#include "CKeyboardNavWrapper.h"

//******************************************************************************
//   Local Macros
//******************************************************************************

//******************************************************************************
//   Local Constants
//******************************************************************************
static const float BORDER(1.f);
static const float POSITION_X(0);
static const float POSITION_Y(22);
static const float WIDTH(CGfx::s_ScreenWidth);
static const float HEIGHT(CGfx::s_ScreenHeight-POSITION_Y);

static const float TEXT_PADDING_X(5.f);
static const float TEXT_PADDING_Y(0.f);
static const float TAB_SPACING(2.f);
static const float SELECTED_TAB_EXTRA_Y(2.f);
static const float TAB_CORNER_RADIUS(5.f);
static const float TAB_BORDER_WIDTH(1.f); //don't mess with this!

//******************************************************************************
//   Static Prototypes
//******************************************************************************

//******************************************************************************
//   Global Variables
//******************************************************************************

//******************************************************************************
//   Static Variables
//******************************************************************************

//******************************************************************************
//   Class Definition
//******************************************************************************

//******************************************************************************
//	
//******************************************************************************
CTabbedWindow::CTabbedWindow() : 
	m_pTabList(NULL),
	m_pSelectedTab(NULL),
	m_maxNameHeight(0.f)
{
	m_pTabList = new TabbedItemList();
}	

//******************************************************************************
//	
//******************************************************************************
CTabbedWindow::~CTabbedWindow()
{
	//loop through the items
	TabbedItemList::iterator it = m_pTabList->begin();

	while (it != m_pTabList->end()) {
		SAFE_DELETE(*it);
		it++;
	}

	SAFE_DELETE(m_pTabList);
}

//******************************************************************************
//	
//******************************************************************************
void CTabbedWindow::Render()
{
	if (m_pSelectedTab != NULL)
	{
		m_pSelectedTab->m_pControlManager->RenderControl();

		// Draw the borders
		// Sort of cheating here, couldn't send bgcolor to none opaque, if we want that functionality
		// need to draw 4 seperate quads
		CGfx::DrawQuad( V2(POSITION_X, POSITION_Y),
						V2(WIDTH, HEIGHT),
						CSkinManager::GetColor( "tabbed_window", "selected_border_color", 0xFF00FF00));
	
		// Draw menu background (overlay over the menu border quad)
		CGfx::DrawQuad( V2(POSITION_X + BORDER, POSITION_Y + BORDER),
						V2(WIDTH - BORDER * 2, HEIGHT - BORDER * 2),
						CSkinManager::GetColor( "tabbed_window", "background_color", 0xFF00FF00));

		m_pSelectedTab->m_pControlManager->RenderControl();

		// Overwrite any controls that appear in the tab space
		CGfx::DrawQuad( V2(POSITION_X, 0),
						V2(WIDTH, POSITION_Y),
						CSkinManager::GetColor( "tabbed_window", "background_color", 0xFF00FF00));

		// Draw the unselected tabs
		V2 tabPosition(0, 0);

		tabPosition.x += TAB_SPACING;
		tabPosition.y = POSITION_Y;

		//draw the unselected tabs
		TabbedItemList::iterator it = m_pTabList->begin();
		while (it != m_pTabList->end())
		{
			if ((*it) != m_pSelectedTab) {
				// Check if next tab is selected
				it++;
				if (it != m_pTabList->end() && *it == m_pSelectedTab) {
					it--;
					tabPosition.x += drawTab((*it)->m_name, tabPosition, false, true, false);
					
					it++;
					tabPosition.x += TAB_SPACING;
					continue;
				}
				it--; //get back to the correct iterator

				// Check if prev tab is selected
				if (it != m_pTabList->begin()) { //if i'm the first one, can't have prev
					it--;
					if (*it == m_pSelectedTab) {
						it++;
						//adjust for wider previous item
						tabPosition.x -= 2*TAB_SPACING;
						tabPosition.x += drawTab((*it)->m_name, tabPosition, false, false, true);

						it++;
						tabPosition.x += TAB_SPACING;
						continue;
					}
					it++; //get back to the correct iterator
				}
				
				// No surrounding tabs selected
				tabPosition.x += drawTab((*it)->m_name, tabPosition, false, false, false);
			} else {
				tabPosition.x += drawTab((*it)->m_name, tabPosition, true, false, false);
			}
			tabPosition.x += TAB_SPACING;
			it++;
		}
	}
}

//******************************************************************************
//	
//******************************************************************************
void CTabbedWindow::Process()
{
	CWindow::Process();

	if (m_pSelectedTab != NULL)
	{
		m_pSelectedTab->m_pControlManager->ProcessControl();
	}
}

//******************************************************************************
//	
//******************************************************************************
bool CTabbedWindow::ProcessInput()
{
	CWindow::ProcessInput();

	TabbedItemList::iterator it;

	if (m_pSelectedTab != NULL) 
	{
		if (!m_pSelectedTab->m_pControlManager->ProcessControlInput()) {
			return false;
		}
		
		// Shift selected tab left one unless already first in list
		if (CKeyboardNavWrapper::PrevTab()) {
			CInput::GetKeyboardCharKey(); // "consume" the key
			it = m_pTabList->begin();
			if (m_pSelectedTab != (*it)) {
				while( it !=  m_pTabList->end() ) {
					if (m_pSelectedTab == (*it)) {
						it--;
						m_pSelectedTab = (*it);
						return false;
					}
					it++;
				}
			}
		}

		// Shift selected tab right one unless already last in list
		if (CKeyboardNavWrapper::NextTab()) {
			CInput::GetKeyboardCharKey(); // "consume" the key
			it = m_pTabList->begin();
			while( it !=  m_pTabList->end() ) {
				if (m_pSelectedTab == (*it)) {
					if (++it != m_pTabList->end()) {
						m_pSelectedTab = (*it);
						return false;
					}
					continue; //already incremented
				}
				it++;
			}
		}
	}
	return true;
}

//******************************************************************************
//	
//******************************************************************************
void CTabbedWindow::addTab(CTabbedWindowItem* tab)
{
	if (m_pSelectedTab == NULL) {
		m_pSelectedTab = tab;
	}
	m_pTabList->push_back(tab);

	// Determine the height of the largest name
	TabbedItemList::iterator it = m_pTabList->begin();

	while (it != m_pTabList->end()) {
		V2 text_size(CFont::GetActiveUiFont()->GetStringSize((*it)->m_name));
		m_maxNameHeight = (text_size.y > m_maxNameHeight) ? text_size.y : m_maxNameHeight;
		it++;
	}
}

//******************************************************************************
//	
//******************************************************************************
float CTabbedWindow::drawTab(CString name, V2 pos, bool selected, bool nextSelected, bool prevSelected)
{
	V2 textSize(CFont::GetActiveUiFont()->GetStringSize(name));
	V2 tabSize(TAB_BORDER_WIDTH*2 + TEXT_PADDING_X*2 + textSize.x, TAB_BORDER_WIDTH + TEXT_PADDING_Y*2 + m_maxNameHeight);

	pos.y -= tabSize.y;

	ARGB color;

	if (selected) {
		color = CSkinManager::GetColor("tabbed_window", "selected_border_color", 0xFF00FF00);
		tabSize.x += 2*TAB_SPACING;
		tabSize.y += SELECTED_TAB_EXTRA_Y;
		pos.x -= TAB_SPACING;
		pos.y -= SELECTED_TAB_EXTRA_Y;

		//Remove border
		CGfx::DrawQuad( V2(pos.x + TAB_BORDER_WIDTH, POSITION_Y),
						V2(tabSize.x - TAB_BORDER_WIDTH, BORDER),
						CSkinManager::GetColor( "tabbed_window", "background_color", 0xFF00FF00));

	} else {
		color = CSkinManager::GetColor("tabbed_window", "border_color", 0xFF00FF00);
	}

	if (!prevSelected) {
		//Draw left vert bar
		CGfx::DrawQuad( V2(pos.x, pos.y + TAB_CORNER_RADIUS),
						V2(TAB_BORDER_WIDTH, tabSize.y - TAB_CORNER_RADIUS),
						color);

		//Draw left corner
		CGfx::DrawArc( pos + V2(TAB_CORNER_RADIUS, TAB_CORNER_RADIUS),
					   TAB_CORNER_RADIUS,
					   G_PI/2,
					   G_PI,
					   color);
	}

	if (!nextSelected) {
		//Draw right vert bar
		CGfx::DrawQuad( V2(pos.x + tabSize.x, pos.y + TAB_CORNER_RADIUS),
						V2(TAB_BORDER_WIDTH, tabSize.y - TAB_CORNER_RADIUS),
						color);

		//Draw right corner
		CGfx::DrawArc( V2(pos.x + tabSize.x - TAB_CORNER_RADIUS + 1, pos.y + TAB_CORNER_RADIUS),
				   TAB_CORNER_RADIUS,
				   0,
				   G_PI/2,
				   color);
	}

	//Draw top bar
	if (prevSelected) {
		CGfx::DrawQuad( V2(pos.x + 1, pos.y),
						V2(tabSize.x - TAB_CORNER_RADIUS, TAB_BORDER_WIDTH),
						color);
	} else if (nextSelected) {
		CGfx::DrawQuad( V2(pos.x + TAB_CORNER_RADIUS, pos.y),
						V2(tabSize.x - TAB_CORNER_RADIUS, TAB_BORDER_WIDTH),
						color);
	} else {
		CGfx::DrawQuad( V2(pos.x + TAB_CORNER_RADIUS, pos.y),
						V2(tabSize.x - TAB_CORNER_RADIUS*2 + 1, TAB_BORDER_WIDTH),
						color);
	}
	
	//Print text
	V2 textPosition(pos.x + tabSize.x/2 - textSize.x/2,
					pos.y + TAB_BORDER_WIDTH + tabSize.y/2 - textSize.y/2);

	CFont::GetActiveUiFont()->Print(name, textPosition, CSkinManager::GetColor( "tabbed_window", "text_color", 0xFF00FF00));

	return tabSize.x;
}

//******************************************************************************
//	
//******************************************************************************
CTabbedWindowItem::CTabbedWindowItem(CString name, CControlManager* const controlManager) :
	m_name(name),
	m_pControlManager(controlManager)
{
}

//******************************************************************************
//	
//******************************************************************************
CTabbedWindowItem::~CTabbedWindowItem() 
{
	SAFE_DELETE(m_pControlManager);
}

//*******************************  END OF FILE  ********************************
