/*******************************************************************************

  PSPoste - Wireless Email Client for the Sony PSP
  Copyright (C) T Swann, Benjamin Biron, Sajeeth Cherian, David Cummings

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

  Module :	CTextureManager.cpp

  Description :	

  Last Modified $Date: 2007/01/03 05:31:14 $

  $Revision: 1.3 $

*******************************************************************************/

//******************************************************************************
//   Include Files
//******************************************************************************
#include "CTextureManager.h"
#include "CFileSystem.h"
#include "CFrameWork.h"
#include "CGfx.h"
#include "CCachedTexture.h"
#include "CCachedFont.h"

//******************************************************************************
//   Local Macros
//******************************************************************************

//******************************************************************************
//   Local Constants
//******************************************************************************

//******************************************************************************
//   Static Prototypes
//******************************************************************************

//******************************************************************************
//   Global Variables
//******************************************************************************

//******************************************************************************
//   Static Variables
//******************************************************************************

//******************************************************************************
//   Class Definition
//******************************************************************************

//******************************************************************************
//******************************************************************************
//
//	CTexture manager implementation
//
//******************************************************************************
//******************************************************************************

//******************************************************************************
//	
//******************************************************************************
CTexture *	CTextureManager::Create( const CString & filename, bool warn, CTexture* destTexture )
{
	CTexture *		p_texture( NULL );
	CFile * const	p_file( CFileSystem::Open( filename, "rb" ) );

	if ( p_file != NULL )
	{
		const CString	extension( p_file->GetExtension() );

		if ( extension == "png" || extension == "PNG" )
		{
			p_texture = ReadPNG( p_file, destTexture );
		} 
		else if ( extension == "bmp" || extension == "BMP" )
		{
			p_texture = ReadBMP( p_file, destTexture );
		}
		else
		{
			TRACE( "Unrecognised image extension %s\n", extension.GetPtr() );
			ASSERT( 0, "" );
		}

		CFileSystem::Close( p_file );
	}

	if ( p_texture == NULL )
	{
		if ( warn == true )
		{
			TRACE( "Failed to load image %s\n", filename.GetPtr() );
			ASSERT( 0, "" );
		}
	}

	return p_texture;
}

//******************************************************************************
//	
//******************************************************************************
CCachedTexture* CTextureManager::CreateCached( const CString & filename )
{
	CCachedTexture* p_texture(new CCachedTexture(filename));
	if (p_texture->isDirty())
	{
		/* Source is newer than cache */
		if (!p_texture->initFromSource())
			SAFE_DELETE(p_texture);
	}
	else if (p_texture->isCacheValid())
	{
		/* Cache actually exists */
		if (!p_texture->initFromCache())
			SAFE_DELETE(p_texture);
	}
	else
	{
		/* Neither source nor cache exist */
		SAFE_DELETE(p_texture);
	}
		
	return p_texture;
}

//******************************************************************************
//	
//******************************************************************************
CCachedFont* CTextureManager::CreateCachedFont( const CString & filename, u32 size )
{
	CCachedFont* p_font(new CCachedFont(filename, size));
	if (p_font->isDirty())
	{
		/* Source is newer than cache */
		if (!p_font->initFromSource())
			SAFE_DELETE(p_font);
	}
	else if (p_font->isCacheValid())
	{
		/* Cache actually exists */
		if (!p_font->initFromCache())
			SAFE_DELETE(p_font);
	}
	else
	{
		/* Neither source nor cache exist */
		SAFE_DELETE(p_font);
	}
		
	return p_font;
}

//******************************************************************************
//	
//******************************************************************************
CFont* CTextureManager::CreateFont( const CString & filename, u32 size, CFont* destFont )
{
	return ReadTTF( filename, size, destFont );
}

//******************************************************************************
//******************************************************************************
//
//	CTexture implementation
//
//******************************************************************************
//******************************************************************************

//******************************************************************************
//	
//******************************************************************************
CTexture::CTexture(CTexture::Format format)
:	m_nWidth( 0 )
,	m_nHeight( 0 )
,	m_nCanvasWidth( 0 )
,	m_nCanvasHeight( 0 )
,	m_pBuffer( NULL )
,	m_pOriginalAddress( NULL )
,	m_pVRAM( NULL )
,	m_pixelComponent( GU_TCC_RGBA )
,	m_texEnv( GU_TFX_MODULATE )
,	m_texMin( GU_LINEAR )
,	m_texMag( GU_LINEAR )
,	m_swizzled( false )
{
	switch (format)
	{
		case mono8:
			m_pixelFormat = GU_PSM_T8;
			m_bpp = 1;
			break;
		case rgba32:
			m_pixelFormat = GU_PSM_8888;
			m_bpp = 4;
			break;
		default:
			CLog::Error("Unknown/unsupported texture format");
			m_pixelFormat = GU_PSM_8888;
			m_bpp = 4;
			break;
	}
}

//******************************************************************************
//
//******************************************************************************
CTexture::~CTexture()
{
	SAFE_RDELETE( m_pOriginalAddress );

	if ( m_pVRAM != NULL )
	{
		// Calling FreeVRAM is a bad idea since it just decrements the "current"
		// VRAM pointer. So if we delete a texture that wasn't the last
		// allocated texture, we're screwing up the last allocated texture(s) by
		// freeing their memory.
		//CGfx::FreeVRAM( m_nCanvasWidth * m_nCanvasHeight * m_bpp );
	}
}

//******************************************************************************
//	
//******************************************************************************
bool	CTexture::Init( u32 width, u32 height )
{
	m_nWidth = width;
	m_nHeight = height;
	m_nCanvasWidth = 1;
	m_nCanvasHeight = 1;

	//
	//	Round the width and height up to a power of 2
	//
	while ( m_nCanvasWidth < width )
	{
		m_nCanvasWidth <<= 1;
	}

	while ( m_nCanvasHeight < height )
	{
		m_nCanvasHeight <<= 1;
	}

	//
	//	Align the memory to 16 bytes
	//
	m_pOriginalAddress = new u8[ ( m_bpp * m_nCanvasWidth * m_nCanvasHeight ) + 16 ];
	ASSERT( m_pOriginalAddress != NULL, "Out of memory!" );
	
	m_pBuffer = reinterpret_cast< u8 * >( ( ( reinterpret_cast< u32 >( m_pOriginalAddress ) + 16 ) & 0xfffffff0 ) );
	memset( m_pBuffer, 0x00, m_bpp * m_nCanvasWidth * m_nCanvasHeight );

	return ( m_pOriginalAddress != NULL );
}

//******************************************************************************
//	
//******************************************************************************
void	CTexture::Save( const CString & file_name )
{
	CFile * const	p_file( CFileSystem::Open( file_name, "wb" ) );

	ASSERT( p_file != NULL, "Failed to open texture file" );

	if ( p_file->Write( m_pBuffer, m_bpp * m_nCanvasWidth * m_nCanvasHeight ) == false )
	{
		ASSERT( 0, "Failed to save texture" );
	}

	CFileSystem::Close( p_file );
}

//******************************************************************************
//
//******************************************************************************
void	CTexture::Upload()
{
	sceKernelDcacheWritebackAll();

	m_pVRAM = CGfx::AllocateVRAM( m_nCanvasWidth * m_nCanvasHeight * m_bpp );

	sceGuStart( GU_DIRECT, CGfx::GetDrawList() );

	sceGuCopyImage( m_pixelFormat, 0, 0, m_nCanvasWidth, m_nCanvasHeight, m_nCanvasWidth, m_pBuffer, 0, 0, m_nCanvasWidth, m_pVRAM );

	sceGuFinish();
	sceGuSync( 0, 0 );
}

//******************************************************************************
//
//******************************************************************************
void CTexture::Swizzle()
{
	char* fromData = (char*)malloc(m_nCanvasWidth * m_nCanvasHeight * m_bpp);
	char* toData = (char*)m_pBuffer;
	memcpy(fromData, toData, m_nCanvasWidth * m_nCanvasHeight * m_bpp);
	int p = m_nCanvasWidth * m_bpp;
	for (unsigned int j = 0; j < m_nCanvasHeight; j += 8) {
		for (int i = 0; i < p; i += 16) {
			char* base = &fromData[i + j*p];
			for (int k = 0; k < 8; k++) {
				memcpy(toData, base, 16);
				toData += 16;
				base += p;
			}
		}
	}
	free(fromData);
	m_swizzled = !m_swizzled;
	
	// We must re-upload the image after swizzling it
	if (m_pVRAM != NULL)
		Upload();
}

//******************************************************************************
//
//******************************************************************************
void CTexture::Optimize()
{
	if (!m_swizzled)
		Swizzle();

	if (m_pVRAM == NULL)
		Upload();
}

//*******************************  END OF FILE  ********************************
