#include "irda_KeyboardLayouts.h"
#include <stdbool.h>
#include <pspkernel.h>
#include <psprtc.h>

#define THINKOUTSIDE_MODEL 1
#define POCKETOP_MODEL 2
#define DEBUG_MODEL 3

#define US_LAYOUT 1
#define FR_LAYOUT 2
#define DE_LAYOUT 3

#define REPEAT_DELAY 500000 // ns
#define REPEAT_RATE 60000 // ns

int keyboard_model;
int keyboard_layout;

#define THINKOUTSIDE_MAX 0x5a
kchar thinkoutside_plain[THINKOUTSIDE_MAX];
kchar thinkoutside_shifted[THINKOUTSIDE_MAX];
kchar thinkoutside_func[THINKOUTSIDE_MAX];
kchar thinkoutside_func2[THINKOUTSIDE_MAX];

bool shift, capslock, ctrl, alt, func, pt_shift, num_shift;
int thinkoutside_state;
bool thinkoutside_isirdaframe;
unsigned char lastraw;
int lastscode;
int lastcode;
u64 nextRepeat;

const unsigned char layout_offsets[] = {
    0x1b, 0x09, 0x0a, 0x0b, 0x0c, 0x0d, 0x0e, 0x3c, 0x3d, 0x3e, 0x3f, 0x38, 0x39, 0x32,
    0x18,    0x11, 0x12, 0x13, 0x14, 0x15, 0x16, 0x44, 0x45, 0x46, 0x47, 0x56,    0x57,
    0x58,      0x03, 0x10, 0x2c, 0x2d, 0x2e, 0x2f, 0x4c, 0x4d, 0x4e, 0x48, 0x59,  0x3a,
    0x1a, 0x08,  0x23,  0x22, 0x17,                0x37, 0x24, 0x51, 0x52, 0x53,  0x50
};



int getKeyboardModel()
{
    return keyboard_model;
}

int getKeyboardLayout()
{
    return keyboard_layout;
}

void setKeyboardModel(int model)
{
    keyboard_model = model;
}

void initThinkoutsideLayout(
    const kchar * layout_plain,
    const kchar * layout_shifted,
    const kchar * layout_func,
    const kchar * layout_func2)
{
 	unsigned int i=0;
    for (i=0; i < sizeof(layout_offsets); i++) {
        int offset=layout_offsets[i];
        thinkoutside_plain[offset] = layout_plain[i];
        thinkoutside_shifted[offset] = layout_shifted[i];
        thinkoutside_func[offset] = layout_func[i];
        thinkoutside_func2[offset] = layout_func2[i];
    }
}

void setKeyboardLayout(int layout)
{
    keyboard_layout= layout;
    if (keyboard_model == THINKOUTSIDE_MODEL) {
        switch (keyboard_layout) {
        case US_LAYOUT:
            initThinkoutsideLayout(layout_us_plain, layout_us_shifted,
                layout_us_func, layout_us_func2);
            break;
        case FR_LAYOUT:
            initThinkoutsideLayout(layout_fr_plain, layout_fr_shifted,
                layout_fr_func, layout_fr_func2);
            break;
        case DE_LAYOUT:
            initThinkoutsideLayout(layout_de_plain, layout_de_shifted,
                layout_de_func, layout_de_func2);
            break;
        }
    }
}

void setKeyboardDefaults()
{
    // defaults
    keyboard_model = THINKOUTSIDE_MODEL;
    setKeyboardLayout(US_LAYOUT);
}

void InitIrdaKeyboard()
{
    setKeyboardDefaults();
    shift = capslock = ctrl = alt = func = pt_shift = num_shift = false;
    thinkoutside_state = 0;
    thinkoutside_isirdaframe = false;
    lastraw = 0;
    lastscode = -1;
    lastcode = 0;
    nextRepeat = 0;
}

int thinkoutside(unsigned char* packet)
{
	// UU VV WW XX YY ZZ
	// UU = 0xFF
	// VV = 0xC0
	// WW = Raw key code, 0x5c means repeat
	// XX, YY = ??
	// ZZ = 0xC1
	unsigned char cmd = packet[2];
	bool keyUpCmd = ((cmd & 0x80) == 0x80);
	unsigned char cook = (cmd & 0x7f);
	u64 curTick;
	
	// Let's ensure this is a valid packet
	if (packet[0] != 0xff || packet[1] != 0xc0 || packet[5] != 0xc1)
		return -1;
		
	// Is this packet a "repeat" packet?
	if (cmd == 0x5c)
	{
		// Note that repeat functionality is now handled by
		// thinkoutside_repeat()
		return -1;
	}
	
	// Is the command useful?
	if (cook > THINKOUTSIDE_MAX)
		return -1;
		
	// From now on, we're dealing with the actual character
	int code = thinkoutside_plain[cook];
	
	// Let's process the modifier keys
	switch (code)
	{
		case KB_COMMAND:
			/* haven't figured out what I want yet */
			break;
		case KB_LSHIFT:
			shift = !keyUpCmd;
			lastcode = 0;
			return 0;
		case KB_RSHIFT:
			shift = !keyUpCmd;
			lastcode = 0;
			return 0;
		case KB_FUNC2:
			pt_shift = !keyUpCmd;
			lastcode = 0;
			return 0;
		case KB_CAPLOCK:
			if (!keyUpCmd) capslock = !capslock; /* toggle */
			lastcode = 0;
			return 0;
		case KB_CTL:
			ctrl = !keyUpCmd;
			lastcode = 0;
			return 0;
		case KB_ALT:
			alt = !keyUpCmd;
			lastcode = 0;
			return 0;
		case KB_FUNC:
			func = !keyUpCmd;
			lastcode = 0;
			return 0;
	}
	
	// Now let's alter the character based on the modifier keys
	if (shift)
	{
		code = thinkoutside_shifted[cook];
	}

	if (capslock)
	{
		/* for capslock -- only shift alpha characters */
		if ((code >= 'a') && (code <= 'z'))
			code = thinkoutside_shifted[cook];
	}

	if (pt_shift)
	{
		code = thinkoutside_func2[cook];
	}
    
    if (alt)
    {
		// Meh
    }

    if (func)
    {
		code = thinkoutside_func[cook];
    }

    if (ctrl)
    {
		if ((code >= 'a') && (code <= 'z')) code = code-'a'+1;
		if ((code >= '[') && (code <= '_')) code = code-'A';
		if ((code >= '2') && (code <= '7')) code = code-('2'-(char)26);
	}
	
	// If it's a keyup event and it corresponds to the last code, we must ensure
	// that last code is set to 0 'cause if not, holding the shift key without
	// any characters will repeat the character even if it's no longer down
	if (keyUpCmd)
	{
		if (code != lastcode)
			return 0; // Don't set the last code!
			
		code = 0;
	}
	
	if (code != lastcode)
	{
		sceRtcGetCurrentTick(&curTick);
		nextRepeat = curTick + REPEAT_DELAY;
	}
	
	lastcode = code;		
	return code;
}

// This function should be called every frame and only when thinkoutside()
// doesn't produce a character
int thinkoutside_repeat()
{
	u64 curTick;
	
	if (nextRepeat == 0)
		return 0;
		
	if (lastcode <= 0)
		return 0;
		
	sceRtcGetCurrentTick(&curTick);
	if (curTick >= nextRepeat)
	{
		nextRepeat = curTick + REPEAT_RATE;
		return lastcode;
	}
	
	return 0;
}

int thinkoutside_symbian(unsigned char raw, bool forcebreak)
{
 	  kchar cook;
    /*
     * ThinkOutside IrDA protocol.  0xffc0 scancode 3bytesofstuff(prob crc).
     *
     * I'm going to wait for 0xffc0 then interpret the scancode and ignore
     * until the next 0xffc0.
     *
     * It's come to my attention that there are some Palm Wireless Keyboards
     * which do not send IrDA framing.  Therefore I have adjusted this to
     * handle both cases.
     */
    switch (thinkoutside_state) {
    case 0:
        /* IrDA frame start */
        if (raw == 0xff) {
            thinkoutside_state = 1;
            thinkoutside_isirdaframe = true;
        } else {
            /* otherwise it's a scancode */
            thinkoutside_isirdaframe = false;
            thinkoutside_state = 2;
            goto thinkoutside_state2; /* Weeeee! */
        }
        return -1;
    case 1:
        if (raw == 0xc0) thinkoutside_state = 2;
        else thinkoutside_state = 0;
        return -1;
    case 2:
thinkoutside_state2:
        if (thinkoutside_isirdaframe) thinkoutside_state = 3;
        else thinkoutside_state = 0;
        thinkoutside_isirdaframe = false; /* reset our flag */
        if ((raw == 0x5c)) { /* repeat indicator */
            /* reset the timer so it doesn't forcebreak on us */
            //iForceBreakTimer->Cancel();
            //iForceBreakTimer->Start(1*1000*1000, 1*1000*1000,
                //TCallBack(ForceBreakTick, this));
            return -1;
            //return lastcode;
        }
        break;
    case 3:
        /*
         * I'm waiting for a 0xc1 which terminates an IrDA frame.  That way
         * I can assume that a 0xff will indicate an IrDA frame next time.
         * If there is a 0xc1 within any CRC then I am screwed.  I didn't
         * manually check them all yet.
         */
        if (raw == 0xc1) thinkoutside_state = 0;
        return -1;
    }
    
    if ((raw&0x80) && (lastscode != -1)) {
        /* send keyup event */
        lastscode = -1;
    }

    cook = (unsigned char)(raw&0x7f);
    if (cook >= THINKOUTSIDE_MAX) return -1;
    int code=thinkoutside_plain[cook];
    int scode=0;
        switch (code) {
    case KB_COMMAND:
        /* haven't figured out what I want yet */
        break;
    case KB_LSHIFT:
        shift = (bool)!(raw&0x80);
        code = 0;
        break;
    case KB_RSHIFT:
        shift = (bool)!(raw&0x80);
        code = 0;
        break;
    case KB_FUNC2:
        pt_shift = (bool)!(raw&0x80);
        code = 0;
        break;
    case KB_CAPLOCK:
        if (!(raw&0x80)) capslock = !capslock; /* toggle */
        code = 0;
        break;
    case KB_CTL:
        ctrl = (bool)!(raw&0x80);
        code = 0;
        break;
    case KB_ALT:
        alt = (bool)!(raw&0x80);
        code = 0;
        break;
    case KB_FUNC:
        func = (bool)!(raw&0x80);
        code = 0;
        break;
    }

    if (raw&0x80) return -1; /* high-bit on is break code */
    if (!code && !scode) return -1;
    
        if (shift) {
        code = thinkoutside_shifted[cook];
        //modifier |= EModifierShift; /* doesn't work right on 3650 */
    }

    if (capslock) {
        /* for capslock -- only shift alpha characters */
        if ((code >= 'a') && (code <= 'z')) code = thinkoutside_shifted[cook];
        //modifier |= EModifierShift; /* doesn't work right on 3650 */
    }

    if (pt_shift) {
        code = thinkoutside_func2[cook];    
        //modifier |= EModifierRightFunc;
    }
    
    if (alt) {
        //modifier |= EModifierAlt;
    }

    if (func) {
        code = thinkoutside_func[cook];
        //modifier |= EModifierLeftFunc;
    }
    /*
    switch (code) {
        case KB_ENTER: code = EKeyEnter; break;
        case KB_BACKSPACE: code = EKeyBackspace; break;
        case KB_UP: code = EKeyUpArrow; break;
        case KB_DOWN: code = EKeyDownArrow; break;
        case KB_LEFT: code = EKeyLeftArrow; break;
        case KB_RIGHT: code = EKeyRightArrow; break;
        case KB_TAB: code = EKeyTab; break;
        case KB_ESC: code = EKeyEscape; break;
        case KB_DELETE: code = EKeyDelete; break;
        case KB_DATE:
            code = 0;
            scode = EStdKeyDevice0;
            break;
        case KB_ADDRESS:
            code = 0;
            scode = EStdKeyDevice1;
            break;
        case KB_TODO:
            code = 0;
            scode = EStdKeyDevice3;
            break;
        case KB_MEMO:
            code = 0;
            scode = EStdKeyApplication0;
            break;
        }*/

    /* make it into control chars if needed */
    if (ctrl) {
        if ((code >= 'a') && (code <= 'z')) code = code-'a'+1;
        if ((code >= '[') && (code <= '_')) code = code-'A';
        if ((code >= '2') && (code <= '7')) code = code-('2'-(char)26);
        //modifier |= EModifierCtrl;
    }
    lastcode = code;
    return code;
    /* send keydown event */
    //aEvent.Set(TRawEvent::EKeyDown, scode);
    //ccoeenv->WsSession().SimulateRawEvent(aEvent);
    //lastscode = scode;

    /* send keypress event */
    //TKeyEvent aKeyEvent;
    //aKeyEvent.iCode = code;
    /* trying to autorepeat a code=0 crashes */
    //if (!code) 
        //aKeyEvent.iModifiers = 0;
    //else
        //aKeyEvent.iModifiers = modifier;
    
    //aKeyEvent.iRepeats = 0;
    //aKeyEvent.iScanCode = scode;
    //ccoeenv->WsSession().SimulateKeyEvent(aKeyEvent);

    /*
     * 1 second timer.  When it expires I will force a keyup.  The timer
     * is reset whenever we receive a repeat signal from the keyboard.
     */
    //iForceBreakTimer->Cancel();
    //iForceBreakTimer->Start(1*1000*1000, 1*1000*1000,
        //TCallBack(ForceBreakTick, this));
}

